from os import path
from tomoscan.esrf import TIFFVolume, MultiTIFFVolume, EDFVolume, JP2KVolume
from tomoscan.esrf.volume.singleframebase import VolumeSingleFrameBase
from ..utils import check_supported, get_num_threads
from ..resources.logger import LoggerOrPrint
from ..io.writer import NXProcessWriter, HSTVolVolume, NXVolVolume
from ..io.utils import convert_dict_values
from .params import files_formats


class WriterManager:
    """
    This class is a wrapper on top of all "writers".
    It will create the right "writer" with all the necessary options, and the histogram writer.

    The layout is the following.

       * Single-big-file volume formats (big-tiff, .vol):
           - no start index
           - everything is increasingly written in one file
       * Multiple-frames per file (HDF5 + master-file):
          - needs a start index (change file_prefix) for each partial file
          - needs a subdirectory for partial files
       * One-file-per-frame (tiff, edf, jp2)
          - start_index

    When saving intermediate steps (eg. sinogram): HDF5 format is always used.
    """

    _overwrite_warned = False
    _writer_classes = {
        "hdf5": NXVolVolume,
        "tiff": TIFFVolume,
        "bigtiff": MultiTIFFVolume,
        "jp2": JP2KVolume,
        "edf": EDFVolume,
        "vol": HSTVolVolume,
    }

    def __init__(
        self,
        output_dir,
        file_prefix,
        file_format="hdf5",
        overwrite=False,
        start_index=0,
        logger=None,
        metadata=None,
        histogram=False,
        extra_options=None,
    ):
        """
        Create a Writer from a set of parameters.

        Parameters
        ----------
        output_dir: str
            Directory where the file(s) will be written.
        file_prefix: str
            File prefix (without leading path)
        start_index: int, optional
            Index to start the files numbering (filename_0123.ext).
            Default is 0.
            Ignored for HDF5 extension.
        logger: nabu.resources.logger.Logger, optional
            Logger object
        metadata: dict, optional
            Metadata, eg. information on various processing steps. For HDF5, it will go to "configuration"
        histogram: bool, optional
            Whether to also write a histogram of data. If set to True, it will configure
            an additional "writer".
        extra_options: dict, optional
            Other advanced options to pass to Writer class.
        """
        self.extra_options = extra_options or {}
        self._set_file_format(file_format)
        self.overwrite = overwrite
        self.start_index = start_index
        self.logger = LoggerOrPrint(logger)
        self.histogram = histogram
        self.output_dir = output_dir
        self.file_prefix = file_prefix
        self.metadata = convert_dict_values(metadata or {}, {None: "None"})
        self._init_writer()
        self._init_histogram_writer()

    def _set_file_format(self, file_format):
        check_supported(file_format, files_formats, "file format")
        self.file_format = files_formats[file_format]
        self._is_bigtiff = file_format in ["tiff", "tif"] and any(
            [self.extra_options.get(opt, False) for opt in ["tiff_single_file", "use_bigtiff"]]
        )
        if self._is_bigtiff:
            self.file_format = "bigtiff"

    @staticmethod
    def get_first_fname(vol_writer):
        if hasattr(vol_writer, "file_path"):
            return path.dirname(vol_writer.file_path)
        dirname = vol_writer.data_url.file_path()
        fname = vol_writer.data_url.data_path().format(
            volume_basename=vol_writer._volume_basename,
            index_zfill6=vol_writer.start_index,
            data_extension=vol_writer.extension or vol_writer.DEFAULT_DATA_EXTENSION,
        )
        return path.join(dirname, fname)

    @staticmethod
    def get_fname(vol_writer):
        if hasattr(vol_writer, "file_path"):
            # several frames per file - return the file itself
            return vol_writer.file_path
        # one file per frame - return the directory
        return vol_writer.data_url.file_path()

    def _init_writer(self):
        self._writer_was_already_initialized = self.extra_options.get("writer_initialized", False)
        if self.file_format in ["tiff", "edf", "jp2", "hdf5"]:
            writer_kwargs = {
                "folder": self.output_dir,
                "volume_basename": self.file_prefix,
                "start_index": self.start_index,
                "overwrite": self.overwrite,
            }
            if self.file_format == "hdf5":
                writer_kwargs["data_path"] = self.metadata.get("entry", "entry")
                writer_kwargs["process_name"] = self.metadata.get("process_name", "reconstruction")
                writer_kwargs["create_subfolder"] = self.extra_options.get("create_subfolder", True)
            elif self.file_format == "jp2":
                writer_kwargs["cratios"] = self.metadata.get("jpeg2000_compression_ratio", None)
                writer_kwargs["clip_values"] = self.metadata.get("float_clip_values", None)
                writer_kwargs["n_threads"] = get_num_threads()
        elif self.file_format in ["vol", "bigtiff"]:
            writer_kwargs = {
                "file_path": path.join(
                    self.output_dir, self.file_prefix + "." + self.file_format.replace("bigtiff", "tiff")
                ),
                "overwrite": self.overwrite,
                "append": self.extra_options.get("single_output_file_initialized", False),
            }
            if self.file_format == "vol":
                writer_kwargs["hst_metadata"] = self.extra_options.get("raw_vol_metadata", {})
        else:
            raise ValueError("Unsupported file format: %s" % self.file_format)
        self._h5_entry = self.metadata.get("entry", "entry")
        self.writer = self._writer_classes[self.file_format](**writer_kwargs)
        self.fname = self.get_fname(self.writer)
        # In certain cases, tomoscan needs to remove any previous existing volume filess
        # and avoid calling 'clean_output_data' when writing downstream (for chunk processing)
        if isinstance(self.writer, VolumeSingleFrameBase):
            self.writer.skip_existing_data_files_removal = self._writer_was_already_initialized
        # ---
        if path.exists(self.fname):
            err = "File already exists: %s" % self.fname
            if self.overwrite:
                if not (self.__class__._overwrite_warned):
                    self.logger.warning(err + ". It will be overwritten as requested in configuration")
                    self.__class__._overwrite_warned = True
            else:
                self.logger.fatal(err)
                raise ValueError(err)

    def _init_histogram_writer(self):
        if not self.histogram:
            return
        separate_histogram_file = not (self.file_format == "hdf5")
        if separate_histogram_file:
            fmode = "w"
            hist_fname = path.join(self.output_dir, "histogram_%05d.hdf5" % self.start_index)
        else:
            fmode = "a"
            hist_fname = self.fname
        # Nabu's original NXProcessWriter has to be used here, as histogram is not 3D
        self.histogram_writer = NXProcessWriter(
            hist_fname,
            entry=self._h5_entry,
            filemode=fmode,
            overwrite=True,
        )

    def write_histogram(self, data, config=None, processing_index=1):
        if not (self.histogram):
            return
        self.histogram_writer.write(
            data,
            "histogram",
            processing_index=processing_index,
            config=config,
            is_frames_stack=False,
            direct_access=False,
        )

    def _write_metadata(self):
        self.writer.metadata = self.metadata
        self.writer.save_metadata()

    def write_data(self, data, metadata=None):
        self.writer.data = data
        if metadata is not None:
            self.writer.metadata = metadata
        self.writer.save()
        # self._write_metadata()



from nabu.stitching.stitcher.pre_processing import PreProcessingStitching
from .dumper import PreProcessingStitchingDumper


class PreProcessingYStitcher(
    PreProcessingStitching,
    dumper_cls=PreProcessingStitchingDumper,
    axis=1,
):

    @property
    def serie_label(self) -> str:
        return "y-serie"



import sys
import logging
import argparse
from typing import Optional

from nabu.app.cli_configs import ReduceDarkFlatConfig
from .utils import parse_params_values
from .. import version

from tomoscan.framereducer.method import ReduceMethod
from tomoscan.scanbase import TomoScanBase
from tomoscan.esrf.scan.edfscan import EDFTomoScan
from tomoscan.factory import Factory
from silx.io.url import DataUrl


def _create_data_urls(output_file: Optional[str], output_data_path: Optional[str], name: str):
    """
    util function to compute reduced Data and metadata url(s)
    This only handle the case of hdf5 outputs
    """
    assert name in ("flats", "darks"), f"name is '{name}'"

    def get_data_paths(data_path: Optional[str]) -> tuple:
        """return (data_path, metadta_path)"""
        if data_path is None:
            return "{entry}/" + name + "/{index}", "{entry}/" + name
        elif not data_path.endswith("/{index}"):
            # we are not expecting useds to provide the index but only upstream part
            return data_path + "/{index}", data_path
        else:
            raise RuntimeError(
                "unhandled use case (/index provided) and don;t know where to set the data and the metadata"
            )

    data_path, metadata_path = get_data_paths(output_data_path)

    output_file = output_file or "{scan_prefix}_" + f"{name}.hdf5"
    data_urls = [
        DataUrl(
            file_path=output_file,
            data_path=data_path,
            scheme="silx",
        ),
    ]

    metadata_urls = [
        DataUrl(
            file_path=output_file,
            data_path=metadata_path,
            scheme="silx",
        ),
    ]
    return data_urls, metadata_urls


def reduce_dark_flat(
    scan: TomoScanBase,
    dark_method: ReduceMethod,
    flat_method: ReduceMethod,
    overwrite: bool = False,
    output_reduced_darks_file: Optional[str] = None,
    output_reduced_darks_data_path: Optional[str] = None,
    output_reduced_flats_file: Optional[str] = None,
    output_reduced_flats_data_path: Optional[str] = None,
) -> int:
    """
    calculation of the darks / flats calling tomoscan utils function
    """
    dark_method = ReduceMethod.from_value(dark_method) if dark_method is not None else None
    flat_method = ReduceMethod.from_value(flat_method) if flat_method is not None else None

    # 1. define url where to save the file
    ## 1.1 for darks
    if dark_method is None:
        reduced_darks_data_urls = ()
        reduced_darks_metadata_urls = ()
    elif output_reduced_darks_file is None and output_reduced_darks_data_path is None:
        # if no settings provided then take the default path (the idea is also to be more robust to future modifications)
        reduced_darks_data_urls = scan.REDUCED_DARKS_DATAURLS
        reduced_darks_metadata_urls = scan.REDUCED_DARKS_METADATAURLS
    elif isinstance(scan, EDFTomoScan):
        # simplification of the equation
        raise ValueError("reduce-dark-flat can only compute create dark-flats at default location for edf")
    else:
        reduced_darks_data_urls, reduced_darks_metadata_urls = _create_data_urls(
            output_file=output_reduced_darks_file,
            output_data_path=output_reduced_darks_data_path,
            name="darks",
        )
    ## 1.2 for flats
    if flat_method is None:
        reduced_flats_data_urls = ()
        reduced_flats_metadata_urls = ()
    elif output_reduced_flats_file is None and output_reduced_flats_data_path is None:
        # if no settings provided then take the default path (the idea is also to be more robust to future modifications)
        reduced_flats_data_urls = scan.REDUCED_FLATS_DATAURLS
        reduced_flats_metadata_urls = scan.REDUCED_FLATS_METADATAURLS
    elif isinstance(scan, EDFTomoScan):
        # simplification of the equation
        raise ValueError("reduce-dark-flat can only compute create dark-flats at default location for edf")
    else:
        reduced_flats_data_urls, reduced_flats_metadata_urls = _create_data_urls(
            output_file=output_reduced_flats_file,
            output_data_path=output_reduced_flats_data_path,
            name="flats",
        )

    # 2. compute and save darks / flats

    success = True
    ## 2.1 handle dark
    if dark_method is not None:
        try:
            reduced_darks, darks_metadata = scan.compute_reduced_darks(
                reduced_method=dark_method,
                overwrite=overwrite,
                return_info=True,
            )
        except Exception as e:
            print(f"failed to create reduced darks. Error is {e}")
            success = False
        else:
            scan.save_reduced_darks(
                darks=reduced_darks,
                darks_infos=darks_metadata,
                output_urls=reduced_darks_data_urls,
                metadata_output_urls=reduced_darks_metadata_urls,
                overwrite=overwrite,
            )

    ## 2.2 handle flats
    if flat_method is not None:
        try:
            reduced_flats, flats_metadata = scan.compute_reduced_flats(
                reduced_method=flat_method,
                overwrite=overwrite,
                return_info=True,
            )
        except Exception as e:
            print(f"failed to create reduced flats. Error is {e}")
            success = False
        else:
            scan.save_reduced_flats(
                flats=reduced_flats,
                flats_infos=flats_metadata,
                output_urls=reduced_flats_data_urls,
                metadata_output_urls=reduced_flats_metadata_urls,
                overwrite=overwrite,
            )

    return success


def main(argv=None):
    """
    Compute reduce dark(s) and flat(s) of a dataset
    """
    if argv is None:
        argv = sys.argv[1:]

    args = parse_params_values(
        ReduceDarkFlatConfig,
        parser_description=main.__doc__,
        program_version="nabu " + version,
        user_args=argv,
    )

    scan = Factory.create_scan_object(args["dataset"], entry=args["entry"])
    exit(
        reduce_dark_flat(
            scan=scan,
            dark_method=args["dark_method"],
            flat_method=args["flat_method"],
            overwrite=args["overwrite"],
            output_reduced_darks_file=args["output_reduced_darks_file"],
            output_reduced_darks_data_path=args["output_reduced_darks_data_path"],
            output_reduced_flats_file=args["output_reduced_flats_file"],
            output_reduced_flats_data_path=args["output_reduced_flats_data_path"],
        )
    )



import numpy
from math import ceil
from typing import Union, Optional
from nabu.stitching.overlap import ImageStichOverlapKernel
from nabu.stitching.frame_composition import FrameComposition
from nabu.stitching.alignment import align_frame, _Alignment


def stitch_raw_frames(
    frames: tuple,
    key_lines: tuple,
    overlap_kernels: Union[ImageStichOverlapKernel, tuple],
    output_dtype: numpy.dtype = numpy.float32,
    check_inputs=True,
    raw_frames_compositions: Optional[FrameComposition] = None,
    overlap_frames_compositions: Optional[FrameComposition] = None,
    return_composition_cls=False,
    alignment: _Alignment = "center",
    pad_mode="constant",
    new_unstitched_axis_size: Optional[int] = None,
) -> numpy.ndarray:
    """
    stitches raw frames (already shifted and flat fielded !!!) together using
    raw stitching (no pixel interpolation, y_overlap_in_px is expected to be a int).
    Sttiching depends on the kernel used.

    It can be done:

    * vertically:

                                    X
        ------------------------------------------------------------------>
        |    --------------
        |    |            |
        |    |  Frame 1   |                           --------------
        |    |            |                           |  Frame 1    |
        |    --------------                           |             |
        Y |                         --> stitching  -->  |~ stitching ~|
        |    --------------                           |             |
        |    |            |                           |  Frame 2    |
        |    |  Frame 2   |                            --------------
        |    |            |
        |    --------------
        \/

    * horizontally:

        ------------------------------------------------------------------>
        |    --------------    --------------                     -----------------------
        |    |            |    |            |                     |         ~ ~         |
        Y |    |  Frame 1   |    |  Frame 2   | --> stitching  -->  | Frame 1 ~ ~ Frame 2 |
        |    |            |    |            |                     |         ~ ~         |
        |    --------------    --------------                     -----------------------
        |
        \/

    returns stitched_projection, raw_img_1, raw_img_2, computed_overlap
    proj_0 and pro_1 are already expected to be in a row. Having stitching_height_in_px in common. At top of proj_0
    and at bottom of proj_1

    :param tuple frames: tuple of 2D numpy array. Expected to be Z up oriented at this stage
    :param tuple key_lines: for each jonction define the two lines to overlaid (from the upper and the lower frames). In the reference where 0 is the bottom line of the image.
    :param overlap_kernels: ZStichOverlapKernel overlap kernel to be used or a list of kernel (one per overlap). Define startegy and overlap heights
    :param numpy.dtype output_dtype: dataset dtype. For now must be provided because flat field corrcetion change data type (numpy.float32 for now)
    :param bool check_inputs: if True will do more test on inputs parameters like checking frame shapes, coherence of the request.. As it can be time consuming it is optional
    :param raw_frames_compositions: pre computed raw frame composition. If not provided will compute them. allow providing it to speed up calculation
    :param overlap_frames_compositions: pre computed stitched frame composition. If not provided will compute them. allow providing it to speed up calculation
    :param bool return_frame_compositions: if False return simply the stitched frames. Else return a tuple with stitching frame and the dictionnary with the composition frames...
    :param alignment: how to align frame if two frames have different size along the unstitched axis
    :param pad_mode: how to pad data for alignment (provided to numpy.pad function)
    :param new_unstitched_axis_size: size of the image along the axis not stitched. So it will be the frame width if the stitching axis is 0 and the frame height if the stitching axis is 1
    """
    if overlap_kernels is None or len(overlap_kernels) == 0:
        raise ValueError("overlap kernels must be provided")

    stitched_axis = overlap_kernels[0].stitched_axis
    unstitched_axis = overlap_kernels[0].unstitched_axis

    if check_inputs:
        # check frames are 2D numpy arrays
        def check_frame(proj):
            if not isinstance(proj, numpy.ndarray) and proj.ndim == 2:
                raise ValueError(f"frames are expected to be 2D numpy array")

        [check_frame(frame) for frame in frames]
        for frame_0, frame_1 in zip(frames[:-1], frames[1:]):
            if not (frame_0.ndim == frame_1.ndim == 2):
                raise ValueError("Frames are expected to be 2D")

        # check there is coherence between overlap kernels and frames
        for frame_0, frame_1, kernel in zip(frames[:-1], frames[1:], overlap_kernels):
            if frame_0.shape[stitched_axis] < kernel.overlap_size:
                raise ValueError(
                    f"frame_0 height ({frame_0.shape[stitched_axis]}) is less than kernel overlap ({kernel.overlap_size})"
                )
            if frame_1.shape[stitched_axis] < kernel.overlap_size:
                raise ValueError(
                    f"frame_1 height ({frame_1.shape[stitched_axis]}) is less than kernel overlap ({kernel.overlap_size})"
                )
        # check key lines are coherent with overlp kernels
        if not len(key_lines) == len(overlap_kernels):
            raise ValueError("we expect to have the same number of key_lines then the number of kernel")
        else:
            for key_line in key_lines:
                for value in key_line:
                    if not isinstance(value, (int, numpy.integer)):
                        raise TypeError(f"key_line is expected to be an integer. {type(key_line)} provided")
                    elif value < 0:
                        raise ValueError(f"key lines are expected to be positive values. Get {value} as key line value")

        # check overlap kernel stitching axis are coherent (for now make sure they are all along the same axis)
        if len(overlap_kernels) > 1:
            for previous_kernel, next_kernel in zip(overlap_kernels[:-1], overlap_kernels[1:]):
                if not isinstance(previous_kernel, ImageStichOverlapKernel):
                    raise TypeError(
                        f"overlap kernels must be instances of {ImageStichOverlapKernel}. Get {type(previous_kernel)}"
                    )
                if not isinstance(next_kernel, ImageStichOverlapKernel):
                    raise TypeError(
                        f"overlap kernels must be instances of {ImageStichOverlapKernel}. Get {type(next_kernel)}"
                    )
                if previous_kernel.stitched_axis != next_kernel.stitched_axis:
                    raise ValueError(
                        "kernels with different stitching axis provided. For now all kernels must have the same stitchign axis"
                    )

    if new_unstitched_axis_size is None:
        new_unstitched_axis_size = max([frame.shape[unstitched_axis] for frame in frames])

    frames = tuple(
        [
            align_frame(
                data=frame,
                alignment=alignment,
                new_aligned_axis_size=new_unstitched_axis_size,
                pad_mode=pad_mode,
                alignment_axis=unstitched_axis,
            )
            for frame in frames
        ]
    )

    # step 1: create numpy array that will contain stitching
    # if raw composition doesn't exists create it
    if raw_frames_compositions is None:
        raw_frames_compositions = FrameComposition.compute_raw_frame_compositions(
            frames=frames,
            overlap_kernels=overlap_kernels,
            key_lines=key_lines,
            stitching_axis=stitched_axis,
        )

    new_stitched_axis_size = raw_frames_compositions.global_end[-1] - raw_frames_compositions.global_start[0]
    if stitched_axis == 0:
        stitched_projection_shape = (
            int(new_stitched_axis_size),
            new_unstitched_axis_size,
        )
    else:
        stitched_projection_shape = (
            new_unstitched_axis_size,
            int(new_stitched_axis_size),
        )

    stitch_array = numpy.empty(stitched_projection_shape, dtype=output_dtype)

    # step 2: set raw data
    # fill stitch array with raw data raw data
    raw_frames_compositions.compose(
        output_frame=stitch_array,
        input_frames=frames,
    )

    # step 3 set stitched data

    # 3.1 create stitched overlaps
    stitched_overlap = []
    for frame_0, frame_1, kernel, key_line in zip(frames[:-1], frames[1:], overlap_kernels, key_lines):
        assert kernel.overlap_size >= 0
        frame_0_overlap, frame_1_overlap = get_overlap_areas(
            upper_frame=frame_0,
            lower_frame=frame_1,
            upper_frame_key_line=key_line[0],
            lower_frame_key_line=key_line[1],
            overlap_size=kernel.overlap_size,
            stitching_axis=stitched_axis,
        )

        assert (
            frame_0_overlap.shape[stitched_axis] == frame_1_overlap.shape[stitched_axis] == kernel.overlap_size
        ), f"{frame_0_overlap.shape[stitched_axis]} == {frame_1_overlap.shape[stitched_axis]} == {kernel.overlap_size}"

        stitched_overlap.append(
            kernel.stitch(
                frame_0_overlap,
                frame_1_overlap,
            )[0]
        )
    # 3.2 fill stitched overlap on output array
    if overlap_frames_compositions is None:
        overlap_frames_compositions = FrameComposition.compute_stitch_frame_composition(
            frames=frames,
            overlap_kernels=overlap_kernels,
            key_lines=key_lines,
            stitching_axis=stitched_axis,
        )
    overlap_frames_compositions.compose(
        output_frame=stitch_array,
        input_frames=stitched_overlap,
    )
    if return_composition_cls:
        return (
            stitch_array,
            {
                "raw_composition": raw_frames_compositions,
                "overlap_composition": overlap_frames_compositions,
            },
        )

    return stitch_array


def get_overlap_areas(
    upper_frame: numpy.ndarray,
    lower_frame: numpy.ndarray,
    upper_frame_key_line: int,
    lower_frame_key_line: int,
    overlap_size: int,
    stitching_axis: int,
):
    """
    return the requested area from lower_frame and upper_frame.

    Lower_frame contains at the end of it the 'real overlap' with the upper_frame.
    Upper_frame contains the 'real overlap' at the end of it.

    For some reason the user can ask the stitching height to be smaller than the `real overlap`.

    Here are some drawing to have a better of view of those regions:

    .. image:: images/stitching/z_stitch_real_overlap.png
        :width: 600

    .. image:: z_stitch_stitch_height.png
        :width: 600
    """
    assert stitching_axis in (0, 1)
    for pf, pn in zip((lower_frame_key_line, upper_frame_key_line), ("lower_frame", "upper_frame")):
        if not isinstance(pf, (int, numpy.number)):
            raise TypeError(f"{pn} is expected to be a number. {type(pf)} provided")
    assert overlap_size >= 0

    lf_start = ceil(lower_frame_key_line - overlap_size / 2)
    lf_end = ceil(lower_frame_key_line + overlap_size / 2)
    uf_start = ceil(upper_frame_key_line - overlap_size / 2)
    uf_end = ceil(upper_frame_key_line + overlap_size / 2)

    lf_start, lf_end = min(lf_start, lf_end), max(lf_start, lf_end)
    uf_start, uf_end = min(uf_start, uf_end), max(uf_start, uf_end)
    if lf_start < 0 or uf_start < 0:
        raise ValueError(
            f"requested overlap ({overlap_size}) is incoherent with key line positions ({lower_frame_key_line}, {upper_frame_key_line}) - expected to be smaller."
        )

    if stitching_axis == 0:
        overlap_upper = upper_frame[uf_start:uf_end]
        overlap_lower = lower_frame[lf_start:lf_end]
    elif stitching_axis == 1:
        overlap_upper = upper_frame[:, uf_start:uf_end]
        overlap_lower = lower_frame[:, lf_start:lf_end]
    else:
        raise NotImplementedError
    if not overlap_upper.shape == overlap_lower.shape:
        # maybe in the future: try to reduce one according to the other ????
        raise RuntimeError(
            f"lower and upper frame have different overlap size ({overlap_upper.shape} vs {overlap_lower.shape})"
        )
    return overlap_upper, overlap_lower



from os.path import dirname
import numpy as np
from ..utils import updiv, get_cuda_srcfile
from ..cuda.utils import __has_pycuda__
from ..misc.utils import ConvolutionInfos
from ..cuda.processing import CudaProcessing

if __has_pycuda__:
    from pycuda.compiler import SourceModule


class Convolution:
    """
    A class for performing convolution on GPU with CUDA, but without using
    textures (unlike for example in ``silx.opencl.convolution``)
    """

    def __init__(self, shape, kernel, axes=None, mode=None, extra_options=None, cuda_options=None):
        """
        Constructor of Cuda Convolution.

        Parameters
        -----------
        shape: tuple
            Shape of the array.
        kernel: array-like
            Convolution kernel (1D, 2D or 3D).
        axes: tuple, optional
            Axes along which the convolution is performed,
            for batched convolutions.
        mode: str, optional
            Boundary handling mode. Available modes are:
               - "reflect": cba|abcd|dcb
               - "nearest": aaa|abcd|ddd
               - "wrap": bcd|abcd|abc
               - "constant": 000|abcd|000

            Default is "reflect".
        extra_options: dict, optional
            Advanced options (dict). Current options are:
               - "allocate_input_array": True
               - "allocate_output_array": True
               - "allocate_tmp_array": True
               - "sourcemodule_kwargs": {}
               - "batch_along_flat_dims": True
        """
        self.cuda = CudaProcessing(**(cuda_options or {}))
        self._configure_extra_options(extra_options)
        self._determine_use_case(shape, kernel, axes)
        self._allocate_memory(mode)
        self._init_kernels()

    def _configure_extra_options(self, extra_options):
        self.extra_options = {
            "allocate_input_array": True,
            "allocate_output_array": True,
            "allocate_tmp_array": True,
            "sourcemodule_kwargs": {},
            "batch_along_flat_dims": True,
        }
        extra_opts = extra_options or {}
        self.extra_options.update(extra_opts)
        self.sourcemodule_kwargs = self.extra_options["sourcemodule_kwargs"]

    def _get_dimensions(self, shape, kernel):
        self.shape = shape
        self.data_ndim = self._check_dimensions(shape=shape, name="Data")
        self.kernel_ndim = self._check_dimensions(arr=kernel, name="Kernel")
        Nx = shape[-1]
        if self.data_ndim >= 2:
            Ny = shape[-2]
        else:
            Ny = 1
        if self.data_ndim >= 3:
            Nz = shape[-3]
        else:
            Nz = 1
        self.Nx = np.int32(Nx)
        self.Ny = np.int32(Ny)
        self.Nz = np.int32(Nz)

    def _determine_use_case(self, shape, kernel, axes):
        """
        Determine the convolution use case from the input/kernel shape, and axes.
        """
        self._get_dimensions(shape, kernel)
        if self.kernel_ndim > self.data_ndim:
            raise ValueError("Kernel dimensions cannot exceed data dimensions")
        data_ndim = self.data_ndim
        kernel_ndim = self.kernel_ndim
        self.kernel = kernel.astype("f")

        convol_infos = ConvolutionInfos()
        k = (data_ndim, kernel_ndim)
        if k not in convol_infos.use_cases:
            raise ValueError(
                "Cannot find a use case for data ndim = %d and kernel ndim = %d" % (data_ndim, kernel_ndim)
            )
        possible_use_cases = convol_infos.use_cases[k]

        # If some dimensions are "flat", make a batched convolution along them
        # Ex. data_dim = (1, Nx) -> batched 1D convolution
        if self.extra_options["batch_along_flat_dims"] and (1 in self.shape):
            axes = tuple([curr_dim for numels, curr_dim in zip(self.shape, range(len(self.shape))) if numels != 1])
        #
        self.use_case_name = None
        for uc_name, uc_params in possible_use_cases.items():
            if axes in convol_infos.allowed_axes[uc_name]:
                self.use_case_name = uc_name
                self.use_case_desc = uc_params["name"]
                self.use_case_kernels = uc_params["kernels"].copy()
        if self.use_case_name is None:
            raise ValueError(
                "Cannot find a use case for data ndim = %d, kernel ndim = %d and axes=%s"
                % (data_ndim, kernel_ndim, str(axes))
            )
        # TODO implement this use case
        if self.use_case_name == "batched_separable_2D_1D_3D":
            raise NotImplementedError("The use case %s is not implemented" % self.use_case_name)
        #
        self.axes = axes
        # Replace "axes=None" with an actual value (except for ND-ND)
        allowed_axes = convol_infos.allowed_axes[self.use_case_name]
        if len(allowed_axes) > 1:
            # The default choice might impact perfs
            self.axes = allowed_axes[0] or allowed_axes[1]
        self.separable = self.use_case_name.startswith("separable")
        self.batched = self.use_case_name.startswith("batched")

    def _allocate_memory(self, mode):
        self.mode = mode or "reflect"
        # The current implementation does not support kernel size bigger than data size,
        # except for mode="nearest"
        for i, dim_size in enumerate(self.shape):
            if min(self.kernel.shape) > dim_size and i in self.axes:
                print(
                    "Warning: kernel support is too large for data dimension %d (%d). Forcing convolution mode to 'nearest'"
                    % (i, dim_size)
                )
                self.mode = "nearest"
        #
        option_array_names = {
            "allocate_input_array": "data_in",
            "allocate_output_array": "data_out",
            "allocate_tmp_array": "data_tmp",
        }
        # Nonseparable transforms do not need tmp array
        if not (self.separable):
            self.extra_options["allocate_tmp_array"] = False
        # Allocate arrays
        for option_name, array_name in option_array_names.items():
            if self.extra_options[option_name]:
                value = self.cuda.allocate_array("value", self.shape, np.float32)
            else:
                value = None
            setattr(self, array_name, value)

        if isinstance(self.kernel, np.ndarray):
            self.d_kernel = self.cuda.to_device("d_kernel", self.kernel)
        else:
            if not (isinstance(self.kernel, self.cuda.array_class)):
                raise ValueError("kernel must be either numpy array or pycuda array")
            self.d_kernel = self.kernel
        self._old_input_ref = None
        self._old_output_ref = None
        self._c_modes_mapping = {
            "periodic": 2,
            "wrap": 2,
            "nearest": 1,
            "replicate": 1,
            "reflect": 0,
            "constant": 3,
        }
        mp = self._c_modes_mapping
        if self.mode.lower() not in mp:
            raise ValueError(
                """
                Mode %s is not available. Available modes are:
                %s
                """
                % (self.mode, str(mp.keys()))
            )
        if self.mode.lower() == "constant":
            raise NotImplementedError("mode='constant' is not implemented yet")
        self._c_conv_mode = mp[self.mode]

    def _init_kernels(self):
        if self.kernel_ndim > 1:
            if np.abs(np.diff(self.kernel.shape)).max() > 0:
                raise NotImplementedError("Non-separable convolution with non-square kernels is not implemented yet")
        # Compile source module
        compile_options = [str("-DUSED_CONV_MODE=%d" % self._c_conv_mode)]
        fname = get_cuda_srcfile("convolution.cu")
        nabu_cuda_dir = dirname(fname)
        include_dirs = [nabu_cuda_dir]
        self.sourcemodule_kwargs["options"] = compile_options
        self.sourcemodule_kwargs["include_dirs"] = include_dirs
        with open(fname) as fid:
            cuda_src = fid.read()
        self._module = SourceModule(cuda_src, **self.sourcemodule_kwargs)  # pylint: disable=E0606
        # Blocks, grid
        self._block_size = {1: (32, 1, 1), 2: (32, 32, 1), 3: (16, 8, 8)}[self.data_ndim]  # TODO tune
        self._n_blocks = tuple([int(updiv(a, b)) for a, b in zip(self.shape[::-1], self._block_size)])
        # Prepare cuda kernel calls
        self._cudakernel_signature = {
            1: "PPPiiii",
            2: "PPPiiiii",
            3: "PPPiiiiii",
        }[self.kernel_ndim]
        self.cuda_kernels = {}
        for axis, kern_name in enumerate(self.use_case_kernels):
            self.cuda_kernels[axis] = self._module.get_function(kern_name)
            self.cuda_kernels[axis].prepare(self._cudakernel_signature)

        # Cuda kernel arguments
        kernel_args = [
            self._n_blocks,
            self._block_size,
            None,
            None,
            self.d_kernel.gpudata,
            np.int32(self.kernel.shape[0]),
            self.Nx,
            self.Ny,
            self.Nz,
        ]
        if self.kernel_ndim == 2:
            kernel_args.insert(5, np.int32(self.kernel.shape[1]))
        if self.kernel_ndim == 3:
            kernel_args.insert(5, np.int32(self.kernel.shape[2]))
            kernel_args.insert(6, np.int32(self.kernel.shape[1]))
        self.kernel_args = tuple(kernel_args)
        # If self.data_tmp is allocated, separable transforms can be performed
        # by a series of batched transforms, without any copy, by swapping refs.
        self.swap_pattern = None
        if self.separable:
            if self.data_tmp is not None:
                self.swap_pattern = {
                    2: [("data_in", "data_tmp"), ("data_tmp", "data_out")],
                    3: [
                        ("data_in", "data_out"),
                        ("data_out", "data_tmp"),
                        ("data_tmp", "data_out"),
                    ],
                }
            else:
                raise NotImplementedError("For now, data_tmp has to be allocated")

    def _get_swapped_arrays(self, i):
        """
        Get the input and output arrays to use when using a "swap pattern".
        Swapping refs enables to avoid copies between temp. array and output.
        For example, a separable 2D->1D convolution on 2D data reads:
          data_tmp = convol(data_input, kernel, axis=1) # step i=0
          data_out = convol(data_tmp, kernel, axis=0) # step i=1

        :param i: current step number of the separable convolution
        """
        n_batchs = len(self.axes)
        in_ref, out_ref = self.swap_pattern[n_batchs][i]
        d_in = getattr(self, in_ref)
        d_out = getattr(self, out_ref)
        return d_in, d_out

    def _configure_kernel_args(self, cuda_kernel_args, input_ref, output_ref):
        # TODO more elegant
        if isinstance(input_ref, self.cuda.array_class):
            input_ref = input_ref.gpudata
        if isinstance(output_ref, self.cuda.array_class):
            output_ref = output_ref.gpudata
        if input_ref is not None or output_ref is not None:
            cuda_kernel_args = list(cuda_kernel_args)
            if input_ref is not None:
                cuda_kernel_args[2] = input_ref
            if output_ref is not None:
                cuda_kernel_args[3] = output_ref
            cuda_kernel_args = tuple(cuda_kernel_args)
        return cuda_kernel_args

    @staticmethod
    def _check_dimensions(arr=None, shape=None, name="", dim_min=1, dim_max=3):
        if shape is not None:
            ndim = len(shape)
        elif arr is not None:
            ndim = arr.ndim
        else:
            raise ValueError("Please provide either arr= or shape=")
        if ndim < dim_min or ndim > dim_max:
            raise ValueError("%s dimensions should be between %d and %d" % (name, dim_min, dim_max))
        return ndim

    def _check_array(self, arr):
        if not (isinstance(arr, self.cuda.array_class) or isinstance(arr, np.ndarray)):
            raise TypeError("Expected either pycuda.gpuarray or numpy.ndarray")
        if arr.dtype != np.float32:
            raise TypeError("Data must be float32")
        if arr.shape != self.shape:
            raise ValueError("Expected data shape = %s" % str(self.shape))

    def _set_arrays(self, array, output=None):
        # Either copy H->D or update references.
        if isinstance(array, np.ndarray):
            self.data_in[:] = array[:]
        else:
            self._old_input_ref = self.data_in
            self.data_in = array
        data_in_ref = self.data_in
        if output is not None:
            if not (isinstance(output, np.ndarray)):
                self._old_output_ref = self.data_out
                self.data_out = output
        # Update Cuda kernel arguments with new array references
        self.kernel_args = self._configure_kernel_args(self.kernel_args, data_in_ref, self.data_out)

    def _separable_convolution(self):
        assert len(self.axes) == len(self.use_case_kernels)
        # Separable: one kernel call per data dimension
        for i, axis in enumerate(self.axes):
            in_ref, out_ref = self._get_swapped_arrays(i)
            self._batched_convolution(axis, input_ref=in_ref, output_ref=out_ref)

    def _batched_convolution(self, axis, input_ref=None, output_ref=None):
        # Batched: one kernel call in total
        cuda_kernel = self.cuda_kernels[axis]
        cuda_kernel_args = self._configure_kernel_args(self.kernel_args, input_ref, output_ref)
        ev = cuda_kernel.prepared_call(*cuda_kernel_args)

    def _nd_convolution(self):
        assert len(self.use_case_kernels) == 1
        cuda_kernel = self._module.get_function(self.use_case_kernels[0])
        ev = cuda_kernel.prepared_call(*self.kernel_args)

    def _recover_arrays_references(self):
        if self._old_input_ref is not None:
            self.data_in = self._old_input_ref
            self._old_input_ref = None
        if self._old_output_ref is not None:
            self.data_out = self._old_output_ref
            self._old_output_ref = None
        self.kernel_args = self._configure_kernel_args(self.kernel_args, self.data_in, self.data_out)

    def _get_output(self, output):
        if output is None:
            res = self.data_out.get()
        else:
            res = output
            if isinstance(output, np.ndarray):
                output[:] = self.data_out[:]
        self._recover_arrays_references()
        return res

    def convolve(self, array, output=None):
        """
        Convolve an array with the class kernel.

        :param array: Input array. Can be numpy.ndarray or pycuda.gpuarray.GPUArray.
        :param output: Output array. Can be numpy.ndarray or pycuda.gpuarray.GPUArray.
        """
        self._check_array(array)
        self._set_arrays(array, output=output)
        if self.axes is not None:
            if self.separable:
                self._separable_convolution()
            elif self.batched:
                assert len(self.axes) == 1
                self._batched_convolution(self.axes[0])
            # else: ND-ND convol
        else:
            # ND-ND convol
            self._nd_convolution()

        res = self._get_output(output)
        return res

    __call__ = convolve



from .reader import *


class ChunkReaderHelical(ChunkReader):
    """implements reading by projections subsets"""

    def _set_subregion(self, sub_region):
        super()._set_subregion(sub_region)
        ###########
        # undo the chun_size setting of the base class
        # to avoid allocation of Tera bytes in the helical case
        self.chunk_shape = (1,) + self.shape

    def set_data_buffer(self, data_buffer, pre_allocate=False):
        if data_buffer is not None:
            # overwrite out_dtype
            self.files_data = data_buffer
            self.out_dtype = data_buffer.dtype
            if data_buffer.shape[1:] != self.shape:
                raise ValueError("Expected shape %s but got %s" % (self.shape, data_buffer.shape))
        if pre_allocate:
            self.files_data = np.zeros(self.chunk_shape, dtype=self.out_dtype)

        if (self.binning is not None) and (np.dtype(self.out_dtype).kind in ["u", "i"]):
            raise ValueError(
                "Output datatype cannot be integer when using binning. Please set the 'convert_float' parameter to True or specify a 'data_buffer'."
            )

    def get_binning(self):
        if self.binning is None:
            return 1, 1
        else:
            return self.binning

    def _load_single(self, sub_total_prange_slice=slice(None, None)):
        if sub_total_prange_slice == slice(None, None):
            sorted_files_indices = self._sorted_files_indices
        else:
            sorted_files_indices = self._sorted_files_indices[sub_total_prange_slice]

        for i, fileidx in enumerate(sorted_files_indices):
            file_url = self.files[fileidx]
            self.files_data[i] = self.get_data(file_url)
            self._fileindex_to_idx[fileidx] = i

    def _apply_subsample_fact(self, t):
        if t is not None:
            t = t * self.dataset_subsampling
        return t

    def _load_multi(self, sub_total_prange_slice=slice(None, None)):
        if sub_total_prange_slice == slice(None, None):
            files_to_be_compacted_dict = self.files
            sorted_files_indices = self._sorted_files_indices
        else:
            if self.dataset_subsampling > 1:
                start, stop, step = list(
                    map(
                        self._apply_subsample_fact,
                        [sub_total_prange_slice.start, sub_total_prange_slice.stop, sub_total_prange_slice.step],
                    )
                )
                sub_total_prange_slice = slice(start, stop, step)

            sorted_files_indices = self._sorted_files_indices[sub_total_prange_slice]
            files_to_be_compacted_dict = dict(
                zip(sorted_files_indices, [self.files[idx] for idx in sorted_files_indices])
            )

        urls_compacted = get_compacted_dataslices(files_to_be_compacted_dict, subsampling=self.dataset_subsampling)
        loaded = {}
        start_idx = 0
        for idx in sorted_files_indices:
            url = urls_compacted[idx]
            url_str = str(url)
            is_loaded = loaded.get(url_str, False)
            if is_loaded:
                continue
            ds = url.data_slice()
            delta_z = ds.stop - ds.start
            if ds.step is not None and ds.step > 1:
                delta_z //= ds.step
            end_idx = start_idx + delta_z
            self.files_data[start_idx:end_idx] = self.get_data(url)
            start_idx += delta_z
            loaded[url_str] = True

    def load_files(self, overwrite: bool = False, sub_total_prange_slice=slice(None, None)):
        """
        Load the files whose links was provided at class instantiation.

        Parameters
        -----------
        overwrite: bool, optional
            Whether to force reloading the files if already loaded.
        """

        if self._loaded and not (overwrite):
            raise ValueError("Radios were already loaded. Call load_files(overwrite=True) to force reloading")
        if self.file_reader.multi_load:
            self._load_multi(sub_total_prange_slice)
        else:
            if self.dataset_subsampling > 1:
                assert (
                    False
                ), " in helica pipeline, load file _load_single has not yet been adapted to angular subsampling "
            self._load_single(sub_total_prange_slice)
        self._loaded = True

    load_data = load_files

    @property
    def data(self):
        return self.files_data



from ..processing.transpose import *
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.misc.transpose has been moved to nabu.processing.transpose", do_print=True, func_name="transpose"
)



from itertools import permutations
import pytest
import numpy as np
from scipy.fft import fftn, ifftn, rfftn, irfftn
from nabu.testutils import generate_tests_scenarios, get_data, get_array_of_given_shape, __do_long_tests__
from nabu.cuda.utils import get_cuda_context, __has_pycuda__
from nabu.processing.fft_cuda import SKCUFFT, VKCUFFT, get_available_fft_implems
from nabu.opencl.utils import __has_pyopencl__, get_opencl_context
from nabu.processing.fft_opencl import VKCLFFT, has_vkfft as has_cl_vkfft
from nabu.processing.fft_base import is_fast_axes

available_cuda_fft = get_available_fft_implems()
__has_vkfft__ = "vkfft" in available_cuda_fft
__has_skcuda__ = "skcuda" in available_cuda_fft


scenarios = {
    "shape": [(256,), (300,), (300, 301), (300, 302)],
    "r2c": [True, False],
    "precision": ["simple"],
    "backend": ["cuda", "opencl"],
}

if __do_long_tests__:
    scenarios["shape"].extend([(307,), (125, 126, 260)])
    scenarios["precision"].append("double")

scenarios = generate_tests_scenarios(scenarios)


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls

    cls.data = get_data("chelsea.npz")["data"]
    cls.abs_tol = {
        "simple": {
            1: 5e-3,
            2: 1.0e0,
            3: 5e2,  # !
        },
        "double": {
            1: 1e-10,
            2: 1e-9,
            3: 1e-7,
        },
    }
    if __has_pycuda__:
        cls.cu_ctx = get_cuda_context(cleanup_at_exit=False)
    if __has_pyopencl__:
        cls.cl_ctx = get_opencl_context("all")
    yield
    if __has_pycuda__:
        cls.cu_ctx.pop()


def _get_fft_cls(backend):
    fft_cls = None
    if backend == "cuda":
        if not (__has_vkfft__ and __has_pycuda__):
            pytest.skip("Need vkfft and pycuda to use VKCUFFT")
        fft_cls = VKCUFFT
    if backend == "opencl":
        if not (has_cl_vkfft() and __has_pyopencl__):
            pytest.skip("Need vkfft and pyopencl to use VKCLFFT")
        fft_cls = VKCLFFT
    return fft_cls


@pytest.mark.usefixtures("bootstrap")
class TestFFT:
    def _get_data_array(self, config):
        r2c = config["r2c"]
        shape = config["shape"]
        precision = config["precision"]
        dtype = {
            True: {"simple": np.float32, "double": np.float64},
            False: {"simple": np.complex64, "double": np.complex128},
        }[r2c][precision]
        data = get_array_of_given_shape(self.data, shape, dtype)
        return data

    @staticmethod
    def check_result(res, ref, config, tol, name=""):
        err_max = np.max(np.abs(res - ref))
        err_msg = "%s FFT(%s, r2c=%s): tol=%.2e, but max error = %.2e" % (
            name,
            str(config["shape"]),
            str(config["r2c"]),
            tol,
            err_max,
        )
        assert np.allclose(res, ref, atol=tol), err_msg

    def _do_fft(self, data, r2c, axes=None, return_fft_obj=False, backend_cls=None):
        ctx = self.cu_ctx if backend_cls.backend == "cuda" else self.cl_ctx
        fft = backend_cls(data.shape, data.dtype, r2c=r2c, axes=axes, ctx=ctx)
        d_data = fft.processing.allocate_array("_data", data.shape, dtype=data.dtype)
        d_data.set(data)
        d_out = fft.fft(d_data)
        res = d_out.get()
        return (res, fft) if return_fft_obj else res

    @staticmethod
    def _do_reference_fft(data, r2c, axes=None):
        ref_fft_func = rfftn if r2c else fftn
        ref = ref_fft_func(data, axes=axes)
        return ref

    @staticmethod
    def _do_reference_ifft(data, r2c, axes=None):
        ref_ifft_func = irfftn if r2c else ifftn
        ref = ref_ifft_func(data, axes=axes)
        return ref

    @pytest.mark.skipif(
        not (__has_skcuda__ and __has_pycuda__), reason="Need pycuda and (scikit-cuda or vkfft) for this test"
    )
    @pytest.mark.parametrize("config", scenarios)
    def test_sckcuda(self, config):
        r2c = config["r2c"]
        shape = config["shape"]
        precision = config["precision"]
        ndim = len(shape)
        if ndim == 3 and not (__do_long_tests__):
            pytest.skip("3D FFTs are done only for long tests - use NABU_LONG_TESTS=1")

        data = self._get_data_array(config)

        res, cufft = self._do_fft(data, r2c, return_fft_obj=True, backend_cls=SKCUFFT)
        ref = self._do_reference_fft(data, r2c)

        tol = self.abs_tol[precision][ndim]
        self.check_result(res, ref, config, tol, name="skcuda")

        # Complex-to-complex can also be performed on real data (as in numpy.fft.fft(real_data))
        if not (r2c):
            res = self._do_fft(data, False, backend_cls=SKCUFFT)
            ref = self._do_reference_fft(data, False)
            self.check_result(res, ref, config, tol, name="skcuda")

        # IFFT
        res = cufft.ifft(cufft.output_fft).get()
        self.check_result(res, data, config, tol, name="skcuda")
        # Perhaps we should also check against numpy/scipy ifft,
        # but it does not yield the good shape for R2C on odd-sized data

    @pytest.mark.skipif(
        not (__has_skcuda__ and __has_pycuda__), reason="Need pycuda and (scikit-cuda or vkfft) for this test"
    )
    @pytest.mark.parametrize("config", scenarios)
    def test_skcuda_batched(self, config):
        shape = config["shape"]
        if len(shape) == 1:
            return
        elif len(shape) == 3 and not (__do_long_tests__):
            pytest.skip("3D FFTs are done only for long tests - use NABU_LONG_TESTS=1")
        r2c = config["r2c"]
        tol = self.abs_tol[config["precision"]][len(shape)]

        data = self._get_data_array(config)

        if data.ndim == 2:
            axes_to_test = [(0,), (1,)]
        elif data.ndim == 3:
            # axes_to_test = [(1, 2), (2, 1), (2,)] # See fft.py: works for C2C but not R2C ?
            axes_to_test = [(2,)]

        for axes in axes_to_test:
            res, cufft = self._do_fft(data, r2c, axes=axes, return_fft_obj=True, backend_cls=SKCUFFT)
            ref = self._do_reference_fft(data, r2c, axes=axes)
            self.check_result(res, ref, config, tol, name="skcuda batched axes=%s" % (str(axes)))
            # IFFT
            res = cufft.ifft(cufft.output_fft).get()
            self.check_result(res, data, config, tol, name="skcuda")

    @pytest.mark.parametrize("config", scenarios)
    def test_vkfft(self, config):
        backend = config["backend"]
        fft_cls = _get_fft_cls(backend)

        r2c = config["r2c"]
        shape = config["shape"]
        precision = config["precision"]
        ndim = len(shape)
        if ndim == 3 and not (__do_long_tests__):
            pytest.skip("3D FFTs are done only for long tests - use NABU_LONG_TESTS=1")

        if ndim >= 2 and r2c and shape[-1] & 1:
            pytest.skip("R2C with odd-sized fast dimension is not supported in VKFFT")

        # FIXME - vkfft + POCL fail for R2C in one dimension
        if config["backend"] == "opencl" and r2c and ndim == 1:
            if self.cl_ctx.devices[0].platform.name.strip().lower() == "portable computing language":
                pytest.skip("Something wrong with vkfft + pocl for R2C 1D")
        # ---

        data = self._get_data_array(config)

        res, fft_obj = self._do_fft(data, r2c, return_fft_obj=True, backend_cls=fft_cls)
        ref = self._do_reference_fft(data, r2c)

        tol = self.abs_tol[precision][ndim]
        self.check_result(res, ref, config, tol, name="vkfft_%s" % backend)

        # Complex-to-complex can also be performed on real data (as in numpy.fft.fft(real_data))
        if not (r2c):
            res = self._do_fft(data, False, backend_cls=fft_cls)
            ref = self._do_reference_fft(data, False)
            self.check_result(res, ref, config, tol, name="vkfft_%s" % backend)

        # IFFT
        res = fft_obj.ifft(fft_obj.output_fft).get()
        self.check_result(res, data, config, tol, name="vkfft_%s" % backend)

    @pytest.mark.parametrize("config", scenarios)
    def test_vkfft_batched(self, config):
        backend = config["backend"]
        fft_cls = _get_fft_cls(backend)
        shape = config["shape"]
        if len(shape) == 1:
            return
        elif len(shape) == 3 and not (__do_long_tests__):
            pytest.skip("3D FFTs are done only for long tests - use NABU_LONG_TESTS=1")
        r2c = config["r2c"]
        tol = self.abs_tol[config["precision"]][len(shape)]

        data = self._get_data_array(config)

        if data.ndim >= 2 and r2c and shape[-1] & 1:
            pytest.skip("R2C with odd-sized fast dimension is not supported in VKFFT")

        # For R2C, only fastest axes are supported by vkfft
        if data.ndim == 2:
            axes_to_test = [(1,)]
        elif data.ndim == 3:
            axes_to_test = [
                (1, 2),
                (2,),
            ]
        for axes in axes_to_test:
            res, cufft = self._do_fft(data, r2c, axes=axes, return_fft_obj=True, backend_cls=fft_cls)
            ref = self._do_reference_fft(data, r2c, axes=axes)
            self.check_result(res, ref, config, tol, name="vkfft_%s batched axes=%s" % (backend, str(axes)))
            # IFFT
            res = cufft.ifft(cufft.output_fft).get()
            self.check_result(res, data, config, tol, name="vkfft_%s" % backend)

    @pytest.mark.skipif(not (__do_long_tests__), reason="Use NABU_LONG_TESTS=1 for this test")
    def test_fast_axes_utility_function(self):
        axes_to_test = {
            2: {
                (0, 1): True,
                (1,): True,
                (-1,): True,
                (-2,): False,
                (0,): False,
            },
            3: {
                (0, 1, 2): True,
                (0, 1): False,
                (1, 2): True,
                (2, 1): True,
                (-2, -1): True,
                (2,): True,
                (-1,): True,
            },
        }
        for ndim, axes_ in axes_to_test.items():
            for axes, is_fast in axes_.items():
                possible_axes = [axes]
                if len(axes) > 1:
                    possible_axes = list(permutations(axes, len(axes)))
                for ax in possible_axes:
                    assert is_fast_axes(ndim, ax) is is_fast



from ast import literal_eval
import numpy as np
from psutil import virtual_memory, cpu_count
from pyunitsystem.metricsystem import MetricSystem
from pyunitsystem.energysystem import EnergySI


def get_values_from_file(fname, n_values=None, shape=None, sep=None, any_size=False):
    """
    Read a text file and scan the values inside.
    This function expects one value per line, or values separated with a separator
    defined with the `sep` parameter.

    Parameters
    ----------
    fname: str
        Path of the text file
    n_values: int, optional
        If set to a value, this function will check that it scans exactly this
        number of values.
        Ignored if `shape` is provided
    shape: tuple, optional
        Generalization of n_values for higher dimensions.
    sep: str, optional
        Separator between values. Default is white space
    any_size: bool, optional
        If set to True, then the parameters 'n_values' and 'shape' are ignored.

    Returns
    --------
    arr: numpy.ndarray
        An array containing the values scanned from the text file
    """
    if not (any_size) and not ((n_values is not None) ^ (shape is not None)):
        raise ValueError("Please provide either n_values or shape")
    arr = np.loadtxt(fname, ndmin=1)
    if (n_values is not None) and (arr.shape[0] != n_values):
        if any_size:
            arr = arr[:n_values]
        else:
            raise ValueError("Expected %d values, but could get %d values" % (n_values, arr.shape[0]))
    if (shape is not None) and (arr.shape != shape):
        if any_size:
            arr = arr[: shape[0], : shape[1]]  # TODO handle more dimensions ?
        else:
            raise ValueError("Expected shape %s, but got shape %s" % (shape, arr.shape))
    return arr


def get_memory_per_node(max_mem, is_percentage=True):
    """
    Get the available memory per node in GB.

    Parameters
    ----------
    max_mem: float
        If is_percentage is False, then number is interpreted as an absolute
        number in GigaBytes.
        Otherwise, it should be a number between 0 and 100 and is interpreted
        as a percentage.
    is_percentage: bool
        A boolean indicating whether the parameter max_mem is to be interpreted
        as a percentage of available system memory.
    """
    sys_avail_mem = virtual_memory().available / 1e9
    if is_percentage:
        return (max_mem / 100.0) * sys_avail_mem
    else:
        return min(max_mem, sys_avail_mem)


def get_threads_per_node(max_threads, is_percentage=True):
    """
    Get the available memory per node in GB.

    Parameters
    ----------
    max_threads: float
        If is_percentage is False, then number is interpreted as an absolute
        number of threads.
        Otherwise, it should be a number between 0 and 100 and is interpreted
        as a percentage.
    is_percentage: bool
        A boolean indicating whether the parameter max_threads is to be interpreted
        as a percentage of available system memory.
    """
    sys_n_threads = cpu_count(logical=True)
    if is_percentage:
        return (max_threads / 100.0) * sys_n_threads
    else:
        return min(max_threads, sys_n_threads)


def extract_parameters(params_str, sep=";"):
    """
    Extract the named parameters from a string.

    Example
    --------
    The function can be used as follows:

    >>> extract_parameters("window_width=None; median_filt_shape=(3,3); padding_mode='wrap'")
    ... {'window_width': None, 'median_filt_shape': (3, 3), 'padding_mode': 'wrap'}
    """
    if params_str in ("", None):
        return {}
    params_list = params_str.strip(sep).split(sep)
    res = {}
    for param_str in params_list:
        param_name, param_val_str = param_str.strip().split("=")
        param_name = param_name.strip()
        param_val_str = param_val_str.strip()
        param_val = literal_eval(param_val_str)
        res[param_name] = param_val
    return res


def compact_parameters(params_dict, sep=";"):
    """
    Compact the parameters from a dict into a  string.

    This is the inverse of extract_parameters. It can be used for example in tomwer to convert
    parameters into a string, for example for cor_options, prior to calling a nabu method which
    is expecting an argument to be in the form of a string containing options.

    Example
    --------
    The function can be used as follows:

    >>> compact_parameters( {"side":"near", "near_pos":300 } )
    ... "side=near; nearpos= 300;"
    """
    if params_dict in ({}, None):
        return ""

    res = ""
    for key, val in params_dict.items():
        res = res + "{key} = {val} " + sep
    return res


def is_hdf5_extension(ext):
    return ext.lower() in ["h5", "hdf5", "nx"]


def get_quantities_and_units(string, sep=";"):
    """
    Return a dictionary with quantities as keys, and values in SI.

    Example
    -------
    get_quantities_and_units("pixel_size = 1.2 um ; distance = 1 m")
    Will return
    {"pixel_size": 1.2e-6, "distance": 1}
    """

    result = {}
    quantities = string.split(sep)
    for quantity in quantities:
        quantity_name, value_and_unit = quantity.split("=")
        quantity_name = quantity_name.strip()
        value_and_unit = value_and_unit.strip()
        value, unit = value_and_unit.split()
        val = float(value)
        # Convert to SI
        try:
            # handle metrics
            conversion_factor = MetricSystem.from_str(unit).value
        except ValueError:
            # handle energies
            conversion_factor = EnergySI.from_str(unit).value / EnergySI.KILOELECTRONVOLT.value

        result[quantity_name] = val * conversion_factor
    return result



from ..processing.unsharp_cuda import *
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.misc.unsharp_cuda has been moved to nabu.processing.unsharp_cuda", do_print=True, func_name="unsharp_cuda"
)



import pyopencl as cl
from ..utils import get_opencl_srcfile
from ..opencl.processing import OpenCLProcessing
from ..opencl.kernel import OpenCLKernel
from ..opencl.utils import allocate_texture, check_textures_availability, copy_to_texture
from .filtering_opencl import OpenCLSinoFilter
from .sinogram_opencl import OpenCLSinoMult
from .fbp_base import BackprojectorBase


class OpenCLBackprojector(BackprojectorBase):
    default_extra_options = {**BackprojectorBase.default_extra_options, "use_textures": True}
    backend = "opencl"
    kernel_filename = "backproj.cl"
    backend_processing_class = OpenCLProcessing
    SinoFilterClass = OpenCLSinoFilter
    SinoMultClass = OpenCLSinoMult

    def _check_textures_availability(self):
        self._use_textures = self.extra_options.get("use_textures", True) and check_textures_availability(
            self._processing.ctx
        )

    def _get_kernel_options(self):
        super()._get_kernel_options()
        self._kernel_options.update(
            {
                "file_name": get_opencl_srcfile(self.kernel_filename),
            }
        )

    def _prepare_kernel_args(self):
        super()._prepare_kernel_args()
        block = self.kern_proj_kwargs.pop("block")
        local_size = block
        grid = self.kern_proj_kwargs.pop("grid")
        global_size = (grid[0] * block[0], grid[1] * block[1])
        # global_size = (updiv(self.n_x, 2), updiv(self.n_y, 2))
        self.kern_proj_args.insert(0, self._processing.queue)  # OpenCLProcessing.__call__ expects first arg to be queue
        self.kern_proj_kwargs.update(
            {
                "global_size": global_size,
                "local_size": local_size,
            }
        )

    def _prepare_textures(self):
        if self._use_textures:
            d_sino_ref = self.d_sino_tex = allocate_texture(self._processing.ctx, self.sino_shape)
            self._kernel_options["sourcemodule_options"].append("-DUSE_TEXTURES")
        else:
            self._d_sino = self._processing.allocate_array("_d_sino", self.sino_shape)
            d_sino_ref = self._d_sino.data
        self.kern_proj_args[2] = d_sino_ref

    def _compile_kernels(self):
        self._prepare_kernel_args()
        self._prepare_textures()  # has to be done before compilation for OpenCL (to pass -DUSE_TEXTURES)
        self.kern_proj_args.append(cl.LocalMemory(self._kernel_options["shared_size"]))
        self.gpu_projector = OpenCLKernel(
            self._kernel_options["kernel_name"],
            self._processing.ctx,
            filename=self._kernel_options["file_name"],
            options=self._kernel_options["sourcemodule_options"],
        )
        if self.halftomo and self.rot_center < self.dwidth:
            self.sino_mult = OpenCLSinoMult(self.sino_shape, self.rot_center, ctx=self._processing.ctx)

    def _transfer_to_texture(self, sino, do_checks=True):
        if self._use_textures:
            return copy_to_texture(self._processing.queue, self.d_sino_tex, sino)
        else:
            if id(self._d_sino) == id(sino):
                return
            return cl.enqueue_copy(self._processing.queue, self._d_sino.data, sino.data)

    def _set_kernel_slice_arg(self, d_slice):
        self.kern_proj_args[1] = d_slice



import h5py
import numpy as np
from scipy.special import erf  # pylint: disable=all
import sys
import os
from scipy.ndimage import gaussian_filter
from nxtomo.nxobject.nxdetector import ImageKey
from nabu.resources.nxflatfield import update_dataset_info_flats_darks
from nabu.resources.dataset_analyzer import HDF5DatasetAnalyzer
from ..io.reader import load_images_from_dataurl_dict


def main(argv=None):
    """auxiliary program that can be called to create  default input detector profiles, for nabu helical,
    concerning the weights of the pixels and the "double flat" renormalisation denominator.
    The result is an hdf5 file that can be used as a "processes_file" in the nabu configuration and is used by nabu-helical.
    In particulars cases the user may have fancy masks and correction map and will provide its own processes file,
    and will not need this.

    This code, and in particular the auxiliary function below (that by the way tomwer can use)
    provide a default construction of such maps. The double-flat is set to one and
    the weight is build on the basis of the flat fields   from the dataset with an apodisation on the borders
    which allows to eliminate discontinuities in the contributions from the borders, above and below
    for the z-translations, and on the left or roght border for half-tomo.

    The usage is ::

       nabu-helical-prepare-weights-double nexus_file_name entry_name

    Then the resulting file can be used as processes file in the configuration file of nabu-helical

    """
    if argv is None:
        argv = sys.argv[1:]
    if len(argv) not in [2, 3, 4, 5, 6]:
        message = f""" Usage:
            nabu-helical-prepare-weights-double nexus_file_name entry_name [target_file name [transition_width_vertical [rotation_axis_position [transition_width_vertical]]]]
        """
        print(message)
        sys.exit(1)

    file_name = argv[0]
    if len(os.path.dirname(file_name)) == 0:
        # To make sure that other utility routines can succesfully deal with path within the current directory
        file_name = os.path.join(".", file_name)

    # still tere was some problem with relative path and how they are dealt with in nxtomomill
    # Better to use absolute path
    file_name = os.path.abspath(file_name)

    entry_name = argv[1]
    process_file_name = "double.h5"

    dataset_info = HDF5DatasetAnalyzer(file_name, extra_options={"h5_entry": entry_name})
    update_dataset_info_flats_darks(dataset_info, flatfield_mode=1)

    beam_profile = 0
    my_flats = load_images_from_dataurl_dict(dataset_info.flats)

    for key, flat in my_flats.items():
        beam_profile += flat
    beam_profile = beam_profile / len(list(dataset_info.flats.keys()))

    transition_width_vertical = 50.0

    # the following two line determines the horisontal transition
    # by default a transition on the right ( corresponds to an axis close to the right border)

    rotation_axis_position = beam_profile.shape[1] - 200
    transition_width_horizontal = 100.0

    if len(argv) in [3, 4, 5, 6]:
        process_file_name = argv[2]
        if len(argv) in [4, 5, 6]:
            transition_width_vertical = float(argv[3])
            if len(argv) in [5, 6]:
                rotation_axis_position = (beam_profile.shape[1] - 1) / 2 + float(argv[4])
                if len(argv) in [6]:
                    transition_width_horizontal = 2 * (float(argv[5]))

    create_heli_maps(
        beam_profile,
        process_file_name,
        entry_name,
        transition_width_vertical,
        rotation_axis_position,
        transition_width_horizontal,
    )

    # here we have been called by the cli. The return value 0 means OK
    return 0


def create_heli_maps(
    profile,
    process_file_name,
    entry_name,
    transition_width_vertical,
    rotation_axis_position,
    transition_width_horizontal,
):
    profile = profile / profile.max()
    profile = profile.astype("f")

    profile = gaussian_filter(profile, 10)

    def compute_border_v(L, m, w):
        x = np.arange(L)

        d = (x - L + m).astype("f")
        res_r = (1 - erf(d / w)) / 2

        d = (x - m).astype("f")
        res_l = (1 + erf(d / w)) / 2

        return res_r * res_l

    def compute_border_h(L, r, tw):
        if r > (L - 1) / 2:
            if tw > (L - r):
                tw = max(1.0, L - r)

            m = tw / 2
            w = tw / 5

            x = np.arange(L)

            d = (x - L + m).astype("f")
            res_r = (1 - erf(d / w)) / 2

            return res_r

        else:
            if tw > r:
                tw = max(1.0, r)

            m = tw / 2
            w = tw / 5

            x = np.arange(L)

            d = (x - m).astype("f")
            res_l = (1 + erf(d / w)) / 2

            return res_l

    with h5py.File(process_file_name, mode="a") as fd:
        path_weights = entry_name + "/weights_field/results/data"
        path_double = entry_name + "/double_flatfield/results/data"

        if path_weights in fd:
            del fd[path_weights]
        if path_double in fd:
            del fd[path_double]

        border = compute_border_h(profile.shape[1], rotation_axis_position, transition_width_horizontal)

        border_v = compute_border_v(
            profile.shape[0], round(transition_width_vertical / 2), transition_width_vertical / 5
        )

        fd[path_weights] = (profile * border) * border_v[:, None]
        fd[path_double] = np.ones_like(profile)



from os import path
from time import time
from math import ceil
import numpy as np
from silx.io.url import DataUrl

from ...utils import get_num_threads, remove_items_from_list
from ...resources.logger import LoggerOrPrint
from ...resources.utils import extract_parameters
from ...misc.binning import binning as image_binning
from ...io.reader import EDFStackReader, HDF5Loader, NXTomoReader
from ...preproc.ccd import Log, CCDFilter
from ...preproc.flatfield import FlatField
from ...preproc.distortion import DistortionCorrection
from ...preproc.shift import VerticalShift
from ...preproc.double_flatfield import DoubleFlatField
from ...preproc.phase import PaganinPhaseRetrieval
from ...preproc.ctf import CTFPhaseRetrieval, GeoPars
from ...reconstruction.sinogram import SinoNormalization
from ...reconstruction.filtering import SinoFilter
from ...reconstruction.mlem import __have_corrct__, MLEMReconstructor
from ...processing.rotation import Rotation
from ...reconstruction.rings import MunchDeringer, SinoMeanDeringer, VoDeringer
from ...processing.unsharp import UnsharpMask
from ...processing.histogram import PartialHistogram, hist_as_2Darray
from ..utils import use_options, pipeline_step, get_subregion
from ..reader import bin_image_stack, load_darks_flats
from ..datadump import DataDumpManager
from ..writer import WriterManager

# For now we don't have a plain python/numpy backend for reconstruction
try:
    from ...reconstruction.fbp_opencl import OpenCLBackprojector as Backprojector
except:
    Backprojector = None


class ChunkedPipeline:
    """
    Pipeline for "regular" full-field tomography.
    Data is processed by chunks. A chunk consists in K contiguous lines of all the radios.
    In parallel geometry, a chunk of K radios lines gives K sinograms,
    and equivalently K reconstructed slices.
    """

    backend = "numpy"
    FlatFieldClass = FlatField
    DoubleFlatFieldClass = DoubleFlatField
    CCDCorrectionClass = CCDFilter
    PaganinPhaseRetrievalClass = PaganinPhaseRetrieval
    CTFPhaseRetrievalClass = CTFPhaseRetrieval
    UnsharpMaskClass = UnsharpMask
    ImageRotationClass = Rotation
    VerticalShiftClass = VerticalShift
    MunchDeringerClass = MunchDeringer
    SinoMeanDeringerClass = SinoMeanDeringer
    VoDeringerClass = VoDeringer
    MLogClass = Log
    SinoNormalizationClass = SinoNormalization
    SinoFilterClass = SinoFilter
    FBPClass = Backprojector
    ConebeamClass = None  # unsupported on CPU
    MLEMClass = MLEMReconstructor
    HBPClass = None  # unsupported on CPU
    HistogramClass = PartialHistogram

    _default_extra_options = {}

    # These steps are skipped if the reconstruction is done in two stages.
    # The first stage will skip these steps, and the second stage will do these stages after merging sinograms.
    _reconstruction_steps = ["sino_rings_correction", "reconstruction", "save", "histogram"]

    def __init__(
        self, process_config, chunk_shape, margin=None, logger=None, use_grouped_mode=False, extra_options=None
    ):
        """
        Initialize a "Chunked" pipeline.

        Parameters
        ----------
        processing_config: `ProcessConfig`
            Process configuration.
        chunk_shape: tuple
            Shape of the chunk of data to process, in the form (n_angles, n_z, n_x).
            It has to account for possible cropping of the data, eg. [:, start_z:end_z, start_x:end_x]
            where start_xz and/or end_xz can be other than None.
        margin: tuple, optional
            Margin to use, in the form ((up, down), (left, right)).
            It is used for example when performing phase retrieval or a convolution-like operation:
            some extra data is kept to avoid boundaries issues.
            These boundaries are then discarded: the data volume is eventually cropped as
            `data[U:D, L:R]` where `((U, D), (L, R)) = margin`
            If not provided, no margin is applied.
        logger: `nabu.app.logger.Logger`, optional
            Logger class
        extra_options: dict, optional
            Advanced extra options.


        Notes
        ------

        Using `margin` results in a lesser number of reconstructed slices.
        More specifically, if `margin = (V, H)`, then there will be `delta_z - 2*V`
        reconstructed slices (if the sub-region is in the middle of the volume)
        or `delta_z - V` reconstructed slices (if the sub-region is on top or bottom
        of the volume).
        """
        self.logger = LoggerOrPrint(logger)
        self._set_params(process_config, chunk_shape, extra_options, margin, use_grouped_mode)
        self._init_pipeline()

    def _set_params(self, process_config, chunk_shape, extra_options, margin, use_grouped_mode):
        self.process_config = process_config
        self.dataset_info = self.process_config.dataset_info
        self.processing_steps = self.process_config.processing_steps.copy()
        self.processing_options = self.process_config.processing_options
        self._set_chunk_shape(chunk_shape, use_grouped_mode)
        self.set_subregion(None)
        self._set_margin(margin)
        self._set_extra_options(extra_options)
        self._callbacks = {}
        self._steps_name2component = {}
        self._steps_component2name = {}

    def _set_chunk_shape(self, chunk_shape, use_grouped_mode):
        if len(chunk_shape) != 3:
            raise ValueError("Expected chunk_shape to be a tuple of length 3 in the form (n_z, n_y, n_x)")
        self.chunk_shape = tuple(int(c) for c in chunk_shape)  # cast to int, as numpy.int64 can make pycuda crash
        # TODO: sanity check (eg. compare to size of radios in dataset_info) ?
        # (n_a, n_z, n_x)
        self.radios_shape = (
            ceil(self.chunk_shape[0] / self.process_config.subsampling_factor),
            self.chunk_shape[1] // self.process_config.binning[1],
            self.chunk_shape[2] // self.process_config.binning[0],
        )
        self.n_angles = self.radios_shape[0]
        self.n_slices = self.radios_shape[1]
        self._grouped_processing = False
        if use_grouped_mode or self.chunk_shape[0] < len(self.process_config.rotation_angles(subsampling=False)):
            # TODO allow a certain tolerance in this case ?
            # Reconstruction is still possible (albeit less accurate) if delta is small
            self._grouped_processing = True
            self.logger.debug("Only a subset of angles is processed - Reconstruction will be skipped")
            self.processing_steps, _ = remove_items_from_list(self.processing_steps, self._reconstruction_steps)

    def _set_margin(self, margin):
        if margin is None:
            U, D, L, R = None, None, None, None
        else:
            ((U, D), (L, R)) = get_subregion(margin, ndim=2)

        # Replace "None" with zeros
        U, D, L, R = U or 0, D or 0, L or 0, R or 0

        self.margin = ((U, D), (L, R))
        self._margin_up = U
        self._margin_down = D
        self._margin_left = L
        self._margin_right = R
        self.use_margin = (U + D + L + R) > 0
        self.n_recs = self.chunk_shape[1] - sum(self.margin[0])
        self.radios_cropped_shape = (self.radios_shape[0], self.radios_shape[1] - U - D, self.radios_shape[2] - L - R)
        if self.use_margin:
            self.n_slices -= sum(self.margin[0])

    def set_subregion(self, sub_region):
        """
        Set the data volume sub-region to process.
        Note that processing margin, if any, is contained within the sub-region.

        Parameters
        -----------
        sub_region: tuple
            Data volume sub-region, in the form ((start_a, end_a), (start_z, end_z), (start_x, end_x))
            where the data volume has a layout (angles, Z, X)
        """
        n_angles = self.dataset_info.n_angles
        n_x, n_z = self.dataset_info.radio_dims
        c_a, c_z, c_x = self.chunk_shape
        if sub_region is None:
            # By default, take the sub-region around central slice
            sub_region = (
                (0, c_a),
                (n_z // 2 - c_z // 2, n_z // 2 - c_z // 2 + c_z),
                (n_x // 2 - c_x // 2, n_x // 2 - c_x // 2 + c_x),
            )
        else:
            sub_region = get_subregion(sub_region, ndim=3)
            # check sub-region
            for i, start_end in enumerate(sub_region):
                start, end = start_end
                if start is not None and end is not None:
                    if end - start != self.chunk_shape[i]:
                        raise ValueError(
                            "Invalid (start, end)=(%d, %d) for sub-region (dimension %d): chunk shape is %s, but %d-%d=%d != %d"
                            % (start, end, i, str(self.chunk_shape), end, start, end - start, self.chunk_shape[i])
                        )
            #
        self.logger.debug("Set sub-region to %s" % (str(sub_region)))
        self.sub_region = sub_region
        self._sub_region_xz = sub_region[2] + sub_region[1]
        self._radios_were_cropped = False

    def _set_extra_options(self, extra_options):
        self.extra_options = self._default_extra_options.copy()
        self.extra_options.update(extra_options or {})

    #
    # Callbacks
    #

    def register_callback(self, step_name, callback):
        """
        Register a callback for a pipeline processing step.

        Parameters
        ----------
        step_name: str
            processing step name
        callback: callable
            A function. It will be executed once the processing step `step_name`
            is finished. The function takes only one argument: the class instance.
        """
        if step_name not in self.processing_steps:
            raise ValueError("'%s' is not in processing steps %s" % (step_name, self.processing_steps))
        if step_name in self._callbacks:
            self._callbacks[step_name].append(callback)
        else:
            self._callbacks[step_name] = [callback]

    #
    # Memory management
    #

    def _allocate_array(self, shape, dtype, name=None):
        return np.zeros(shape, dtype=dtype)

    def _allocate_recs(self, ny, nx, n_slices=None):
        n_slices = n_slices or self.n_slices
        self.recs = self._allocate_array((n_slices, ny, nx), "f", name="recs")

    #
    # Runtime attributes
    #

    @property
    def sub_region_xz(self):
        """
        Return the currently processed sub-region in the form
        (start_x, end_x, start_z, end_z)
        """
        return self._sub_region_xz

    @property
    def z_min(self):
        return self._sub_region_xz[2]

    @property
    def sino_shape(self):
        return self.process_config.sino_shape(binning=True, subsampling=True)

    @property
    def sinos_shape(self):
        return (self.n_slices,) + self.sino_shape

    def get_slice_start_index(self):
        return self.z_min + self._margin_up

    #
    # Pipeline initialization
    #

    def _init_pipeline(self):
        self._allocate_radios()
        self._init_data_dump()
        self._init_reader()
        self._init_flatfield()
        self._init_double_flatfield()
        self._init_ccd_corrections()
        self._init_radios_rotation()
        self._init_phase()
        self._init_unsharp()
        self._init_radios_movements()
        self._init_mlog()
        self._init_sino_normalization()
        self._init_sino_rings_correction()
        self._init_reconstruction()
        self._init_histogram()
        self._init_writer()

    def _allocate_radios(self):
        self.radios = np.zeros(self.radios_shape, dtype=np.float32)
        self.data = self.radios  # alias

    def _init_data_dump(self):
        self._resume_from_step = self.processing_options["read_chunk"].get("step_name", None)
        self.datadump_manager = DataDumpManager(
            self.process_config, self.sub_region, margin=self.margin, logger=self.logger
        )
        # When using "grouped processing", sinogram has to be dumped.
        # If it was not specified by user, force sinogram dump
        # Perhaps these lines should be moved directly to DataDumpManager.
        if self._grouped_processing and not self.process_config.dump_sinogram:
            sino_dump_fname = self.process_config.get_save_steps_file("sinogram")
            self.datadump_manager._configure_dump("sinogram", force_dump_to_fname=sino_dump_fname)
            self.logger.debug("Will dump sinogram to %s" % self.datadump_manager.data_dump["sinogram"].fname)

    def _init_reading_processing_function(self):
        # Some processing may be applied directly when reading data (eg. distortion correction, binning, ...)
        # Configure it here
        self._reader_processing_function = None
        self._reader_processing_function_args = None
        self._reader_processing_function_kwargs = None
        self._ff_processing_function = None
        self._ff_processing_function_args = None
        if self.process_config.binning is None or self.process_config.binning == (1, 1):
            return
        if self.dataset_info.kind == "nx":
            self._reader_processing_function = bin_image_stack
            self._reader_processing_function_kwargs = {
                "binning_factor": self.process_config.binning[::-1],
                "num_threads": get_num_threads(),
            }
        else:
            self._reader_processing_function = image_binning
            self._reader_processing_function_args = [self.process_config.binning[::-1]]
        # flat-field is read image-wise
        self._ff_processing_function = image_binning
        self._ff_processing_function_args = [self.process_config.binning[::-1]]

    @use_options("read_chunk", "chunk_reader")
    def _init_reader(self):
        options = self.processing_options["read_chunk"]
        process_file = options.get("process_file", None)
        if process_file is None:  # Standard case - start pipeline from raw data
            self._init_reading_processing_function()

            subs_angles = None
            subs_z = None
            subs_x = None
            if self.process_config.subsampling_factor:
                subs_angles = self.process_config.subsampling_factor
            reader_sub_region = (
                slice(*(self.sub_region[0]) + ((subs_angles,) if subs_angles else ())),
                slice(*(self.sub_region[1]) + ((subs_z,) if subs_z else ())),
                slice(*(self.sub_region[2]) + ((subs_x,) if subs_x else ())),
            )

            other_reader_kwargs = {
                "output_dtype": np.float32,
                "processing_func": self._reader_processing_function,
                "processing_func_args": self._reader_processing_function_args,
                "processing_func_kwargs": self._reader_processing_function_kwargs,
            }

            if self.dataset_info.kind == "nx":
                self.chunk_reader = NXTomoReader(
                    self.dataset_info.dataset_hdf5_url.file_path(),
                    self.dataset_info.dataset_hdf5_url.data_path(),
                    sub_region=reader_sub_region,
                    image_key=0,
                    **other_reader_kwargs,
                )
            elif self.dataset_info.kind == "edf":
                files = [
                    self.dataset_info.projections[k].file_path() for k in sorted(self.dataset_info.projections.keys())
                ]
                self.chunk_reader = EDFStackReader(
                    files,
                    sub_region=reader_sub_region,
                    n_reading_threads=max(1, get_num_threads() // 2),
                    **other_reader_kwargs,
                )
        else:
            # Resume pipeline from dumped intermediate step
            self.chunk_reader = HDF5Loader(
                process_file,
                options["process_h5_path"],
                sub_region=self.datadump_manager.get_read_dump_subregion(),
                data_buffer=self.radios,
                pre_allocate=False,
            )
            self._resume_from_step = options["step_name"]
            self.logger.debug(
                "Load subregion %s from file %s" % (str(self.chunk_reader.sub_region), self.chunk_reader.fname)
            )

    @use_options("flatfield", "flatfield")
    def _init_flatfield(self):
        self._ff_options = self.processing_options["flatfield"].copy()

        # This won't work when resuming from a step (i.e before FF), because we rely on H5Loader()
        # which re-compacts the data. When data is re-compacted, we have to know the original radios positions.
        # These positions can be saved in the "file_dump" metadata, but it is not loaded for now
        # (the process_config object is re-built from scratch every time)
        self._ff_options["projs_indices"] = self.chunk_reader.get_frames_indices()

        if self._ff_options.get("normalize_srcurrent", False):
            a_start_idx, a_end_idx = self.sub_region[0]
            subs = self.process_config.subsampling_factor
            self._ff_options["radios_srcurrent"] = self._ff_options["radios_srcurrent"][a_start_idx:a_end_idx:subs]

        distortion_correction = None
        if self._ff_options["do_flat_distortion"]:
            self.logger.info("Flats distortion correction will be applied")
            self.FlatFieldClass = FlatField  # no GPU implementation available, force this backend
            estimation_kwargs = {}
            estimation_kwargs.update(self._ff_options["flat_distortion_params"])
            estimation_kwargs["logger"] = self.logger
            distortion_correction = DistortionCorrection(
                estimation_method="fft-correlation", estimation_kwargs=estimation_kwargs, correction_method="interpn"
            )

        # Reduced darks/flats are loaded, but we have to crop them on the current sub-region
        # and possibly do apply some pre-processing (binning, distortion correction, ...)
        darks_flats = load_darks_flats(
            self.dataset_info,
            self.sub_region[1:],
            processing_func=self._ff_processing_function,
            processing_func_args=self._ff_processing_function_args,
        )

        # FlatField parameter "radios_indices" must account for subsampling
        self.flatfield = self.FlatFieldClass(
            self.radios_shape,
            flats=darks_flats["flats"],
            darks=darks_flats["darks"],
            radios_indices=self._ff_options["projs_indices"],
            interpolation="linear",
            distortion_correction=distortion_correction,
            radios_srcurrent=self._ff_options["radios_srcurrent"],
            flats_srcurrent=self._ff_options["flats_srcurrent"],
        )

    @use_options("double_flatfield", "double_flatfield")
    def _init_double_flatfield(self):
        options = self.processing_options["double_flatfield"]
        avg_is_on_log = options["sigma"] is not None
        result_url = None
        if options["processes_file"] not in (None, ""):
            result_url = DataUrl(
                file_path=options["processes_file"],
                data_path=(self.dataset_info.hdf5_entry or "entry") + "/double_flatfield/results/data",
            )
            self.logger.info("Loading double flatfield from %s" % result_url.file_path())
        if (self.n_angles < self.process_config.n_angles(subsampling=True)) and result_url is None:
            raise ValueError(
                "Cannot use double-flatfield when processing subset of radios. Please use the 'nabu-double-flatfield' command"
            )
        self.double_flatfield = self.DoubleFlatFieldClass(
            self.radios_shape,
            result_url=result_url,
            sub_region=self.sub_region[1:],
            input_is_mlog=False,
            output_is_mlog=False,
            average_is_on_log=avg_is_on_log,
            sigma_filter=options["sigma"],
            log_clip_min=options["log_min_clip"],
            log_clip_max=options["log_max_clip"],
        )

    @use_options("ccd_correction", "ccd_correction")
    def _init_ccd_corrections(self):
        options = self.processing_options["ccd_correction"]
        self.ccd_correction = self.CCDCorrectionClass(
            self.radios_shape[1:], median_clip_thresh=options["median_clip_thresh"]
        )

    @use_options("tilt_correction", "projs_rot")
    def _init_radios_rotation(self):
        options = self.processing_options["tilt_correction"]
        center = options["center"]
        if center is None:
            nz, nx = self.radios_shape[1:]  # after binning
            center_x = self.process_config.rotation_axis_position(binning=True)
            center_z = nz / 2 - 0.5
            center = (center_x, center_z)
        center = (center[0], center[1] - self.z_min)
        self.projs_rot = self.ImageRotationClass(
            self.radios_shape[1:], options["angle"], center=center, mode="edge", reshape=False
        )
        self._tmp_rotated_radio = self._allocate_array(self.radios_shape[1:], "f", name="tmp_rotated_radio")

    @use_options("radios_movements", "radios_movements")
    def _init_radios_movements(self):
        options = self.processing_options["radios_movements"]
        self._vertical_shifts = options["translation_movements"][:, 1]
        self.radios_movements = self.VerticalShiftClass(self.radios.shape, self._vertical_shifts)

    @use_options("phase", "phase_retrieval")
    def _init_phase(self):
        options = self.processing_options["phase"]
        if options["method"] == "CTF":
            translations_vh = getattr(self.dataset_info, "ctf_translations", None)
            geo_pars_params = options["ctf_geo_pars"].copy()
            geo_pars_params["logger"] = self.logger
            geo_pars = GeoPars(**geo_pars_params)
            self.phase_retrieval = self.CTFPhaseRetrievalClass(
                self.radios_shape[1:],
                geo_pars,
                options["delta_beta"],
                lim1=options["ctf_lim1"],
                lim2=options["ctf_lim2"],
                logger=self.logger,
                fft_num_threads=None,  # TODO tune in advanced params of nabu config file
                use_rfft=True,
                normalize_by_mean=options["ctf_normalize_by_mean"],
                translation_vh=translations_vh,
            )
        else:
            self.phase_retrieval = self.PaganinPhaseRetrievalClass(
                self.radios_shape[1:],
                distance=options["distance_m"],
                energy=options["energy_kev"],
                delta_beta=options["delta_beta"],
                pixel_size=options["pixel_size_m"],
                padding=options["padding_type"],
                # TODO tune in advanced params of nabu config file
                fft_num_threads=None,
            )

    @use_options("unsharp_mask", "unsharp_mask")
    def _init_unsharp(self):
        options = self.processing_options["unsharp_mask"]
        self.unsharp_mask = self.UnsharpMaskClass(
            self.radios_shape[1:],
            options["unsharp_sigma"],
            options["unsharp_coeff"],
            mode="reflect",
            method=options["unsharp_method"],
        )

    @use_options("take_log", "mlog")
    def _init_mlog(self):
        options = self.processing_options["take_log"]
        self.mlog = self.MLogClass(
            self.radios_shape, clip_min=options["log_min_clip"], clip_max=options["log_max_clip"]
        )

    @use_options("sino_normalization", "sino_normalization")
    def _init_sino_normalization(self):
        options = self.processing_options["sino_normalization"]
        self.sino_normalization = self.SinoNormalizationClass(
            kind=options["method"],
            radios_shape=self.radios_cropped_shape,
            normalization_array=options["normalization_array"],
        )

    @use_options("sino_rings_correction", "sino_deringer")
    def _init_sino_rings_correction(self):
        n_a, n_z, n_x = self.radios_cropped_shape
        sinos_shape = (n_z, n_a, n_x)
        options = self.processing_options["sino_rings_correction"]

        destriper_params = extract_parameters(options["user_options"])
        if options["method"] == "munch":
            # TODO MunchDeringer does not have an API consistent with the other deringers
            fw_sigma = destriper_params.pop("sigma", 1.0)
            self.sino_deringer = self.MunchDeringerClass(fw_sigma, sinos_shape, **destriper_params)
        elif options["method"] == "vo":
            self.sino_deringer = self.VoDeringerClass(sinos_shape, **destriper_params)
        elif options["method"] == "mean-subtraction":
            self.sino_deringer = self.SinoMeanDeringerClass(
                sinos_shape, mode="subtract", fft_num_threads=None, **destriper_params
            )
        elif options["method"] == "mean-division":
            self.sino_deringer = self.SinoMeanDeringerClass(
                sinos_shape, mode="divide", fft_num_threads=None, **destriper_params
            )

    @use_options("reconstruction", "reconstruction")
    def _init_reconstruction(self):
        options = self.processing_options["reconstruction"]
        if options["method"] == "FBP" and self.FBPClass is None:
            raise ValueError("No usable FBP module was found")
        if options["method"] == "cone" and self.ConebeamClass is None:
            raise ValueError("No usable cone-beam module was found")
        if options["method"] == "mlem" and self.MLEMClass is None:
            raise ValueError("No usable MLEM module was found.")

        n_slices = self.n_slices
        if options["method"] in ["FBP", "HBP"]:  # both have the same API
            rec_cls = self.HBPClass if options["method"] == "HBP" else self.FBPClass
            self.reconstruction = rec_cls(
                self.sinos_shape[1:],
                angles=options["angles"],
                rot_center=options["rotation_axis_position"],
                filter_name=options["fbp_filter_type"] or "none",
                halftomo=options["enable_halftomo"],
                slice_roi=self.process_config.rec_roi,
                padding_mode=options["padding_type"],
                extra_options={
                    "scale_factor": 1.0 / options["voxel_size_cm"][0],
                    "axis_correction": options["axis_correction"],
                    "centered_axis": options["centered_axis"],
                    "clip_outer_circle": options["clip_outer_circle"],
                    "outer_circle_value": options["outer_circle_value"],
                    "filter_cutoff": options["fbp_filter_cutoff"],
                    "hbp_legs": options["hbp_legs"],
                    "hbp_reduction_steps": options["hbp_reduction_steps"],
                },
            )

        if options["method"] == "cone":
            n_slices = self.n_slices + sum(self.margin[0])
            # For numerical stability, normalize all lengths with respect to detector pixel size
            pixel_size_m = self.dataset_info.pixel_size * 1e-6
            source_sample_dist = options["source_sample_dist"] / pixel_size_m
            sample_detector_dist = options["sample_detector_dist"] / pixel_size_m
            self.reconstruction = self.ConebeamClass(  # pylint: disable=E1102
                (self.radios_shape[1],) + self.sino_shape,
                source_sample_dist,
                sample_detector_dist,
                angles=-options["angles"],
                rot_center=options["rotation_axis_position"],
                pixel_size=1,
                padding_mode=options["padding_type"],
                slice_roi=self.process_config.rec_roi,
                extra_options={
                    "scale_factor": 1.0 / options["voxel_size_cm"][0],
                    "axis_correction": -options["axis_correction"] if options["axis_correction"] is not None else None,
                    "clip_outer_circle": options["clip_outer_circle"],
                    "outer_circle_value": options["outer_circle_value"],
                    "filter_cutoff": options["fbp_filter_cutoff"],
                },
            )

        if options["method"] == "mlem" and options["implementation"] in (None, "corrct"):
            self.reconstruction = self.MLEMClass(  # pylint: disable=E1102
                (self.radios_shape[1],) + self.sino_shape,
                angles_rad=-options["angles"],  # WARNING: mind the sign...
                shifts_uv=self.dataset_info.translations,  # In config file, one line per proj, each line is (tu,tv). Corrct expects one col per proj and (tv,tu).
                cor=options["rotation_axis_position"],
                n_iterations=options["iterations"],
                extra_options={
                    "compute_shifts": False,
                    "tomo_consistency": False,
                    "v_min_for_v_shifts": 0,
                    "v_max_for_v_shifts": None,
                    "v_min_for_u_shifts": 0,
                    "v_max_for_u_shifts": None,
                },
            )

        self._allocate_recs(*self.process_config.rec_shape, n_slices=n_slices)
        n_a, _, n_x = self.radios_cropped_shape
        self._tmp_sino = self._allocate_array((n_a, n_x), "f", name="tmp_sino")

    @use_options("histogram", "histogram")
    def _init_histogram(self):
        options = self.processing_options["histogram"]
        self.histogram = self.HistogramClass(method="fixed_bins_number", num_bins=options["histogram_bins"])

    @use_options("save", "writer")
    def _init_writer(self, **extra_options):
        options = self.processing_options["save"]
        metadata = {
            "process_name": "reconstruction",
            "processing_index": 0,
            # TODO this one takes too much time to write, not useful for partial files
            # "processing_options": self.processing_options,
            #
            "nabu_config": self.process_config.nabu_config,
            "entry": getattr(self.dataset_info.dataset_scanner, "entry", "entry"),
        }
        writer_extra_options = {
            "jpeg2000_compression_ratio": options["jpeg2000_compression_ratio"],
            "float_clip_values": options["float_clip_values"],
            "tiff_single_file": options.get("tiff_single_file", False),
            "single_output_file_initialized": getattr(
                self.process_config, "single_output_file_initialized", False
            ),  # COMPAT.
            "writer_initialized": getattr(self.process_config, "_writer_initialized", False),
            "raw_vol_metadata": {"voxelSize": self.dataset_info.pixel_size},  # legacy...
        }
        writer_extra_options.update(extra_options)
        self.writer = WriterManager(
            options["location"],
            options["file_prefix"],
            file_format=options["file_format"],
            overwrite=options["overwrite"],
            start_index=self.get_slice_start_index(),
            logger=self.logger,
            metadata=metadata,
            histogram=("histogram" in self.processing_steps),
            extra_options=writer_extra_options,
        )

    #
    # Pipeline execution
    #

    @pipeline_step("chunk_reader", "Reading data")
    def _read_data(self):
        self.logger.debug("Region = %s" % str(self.sub_region))
        t0 = time()
        self.chunk_reader.load_data(output=self.radios)
        el = time() - t0
        self.logger.info("Read subvolume %s in %.2f s" % (str(self.radios.shape), el))

    @pipeline_step("flatfield", "Applying flat-field")
    def _flatfield(self):
        self.flatfield.normalize_radios(self.radios)

    @pipeline_step("double_flatfield", "Applying double flat-field")
    def _double_flatfield(self, radios=None):
        if radios is None:
            radios = self.radios
        self.double_flatfield.apply_double_flatfield(radios)

    @pipeline_step("ccd_correction", "Applying CCD corrections")
    def _ccd_corrections(self, radios=None):
        if radios is None:
            radios = self.radios
        _tmp_radio = self._allocate_array(radios.shape[1:], "f", name="tmp_ccdcorr_radio")
        for i in range(radios.shape[0]):
            self.ccd_correction.median_clip_correction(radios[i], output=_tmp_radio)
            radios[i][:] = _tmp_radio[:]

    @pipeline_step("projs_rot", "Rotating projections")
    def _rotate_projections(self, radios=None):
        if radios is None:
            radios = self.radios
        tmp_radio = self._tmp_rotated_radio
        for i in range(radios.shape[0]):
            self.projs_rot.rotate(radios[i], output=tmp_radio)
            radios[i][:] = tmp_radio[:]

    @pipeline_step("phase_retrieval", "Performing phase retrieval")
    def _retrieve_phase(self):
        for i in range(self.radios.shape[0]):
            self.phase_retrieval.retrieve_phase(self.radios[i], output=self.radios[i])

    @pipeline_step("unsharp_mask", "Performing unsharp mask")
    def _apply_unsharp(self):
        for i in range(self.radios.shape[0]):
            self.radios[i] = self.unsharp_mask.unsharp(self.radios[i])

    @pipeline_step("mlog", "Taking logarithm")
    def _take_log(self):
        self.mlog.take_logarithm(self.radios)

    @pipeline_step("radios_movements", "Applying radios movements")
    def _radios_movements(self, radios=None):
        if radios is None:
            radios = self.radios
        self.radios_movements.apply_vertical_shifts(radios, list(range(radios.shape[0])))

    def _crop_radios(self):
        if self.use_margin:
            self._orig_radios = self.radios
            if self.processing_options.get("reconstruction", {}).get("method", None) in ("cone",):
                return
            ((U, D), (L, R)) = self.margin
            self.logger.debug(
                "Cropping radios from %s to %s" % (str(self.radios_shape), str(self.radios_cropped_shape))
            )
            U, D, L, R = U or None, -D or None, L or None, -R or None
            self.radios = self.radios[:, U:D, L:R]  # view
            self._radios_were_cropped = True

    @pipeline_step("sino_normalization", "Normalizing sinograms")
    def _normalize_sinos(self, radios=None):
        if radios is None:
            radios = self.radios
        sinos = radios.transpose((1, 0, 2))
        self.sino_normalization.normalize(sinos)

    def _dump_sinogram(self):
        if self.datadump_manager is not None:
            self.datadump_manager.dump_data_to_file("sinogram", self.radios)

    @pipeline_step("sino_deringer", "Removing rings on sinograms")
    def _destripe_sinos(self):
        sinos = np.rollaxis(self.radios, 1, 0)  # view
        self.sino_deringer.remove_rings(sinos)  # TODO check it works with non-contiguous view

    @pipeline_step("reconstruction", "Reconstruction")
    def _reconstruct(self):
        """
        Reconstruction for parallel geometry.
        For each target slice: get the corresponding sinogram, apply some processing, then reconstruct
        """
        options = self.processing_options["reconstruction"]
        if options["method"] == "cone":
            self._reconstruct_cone()
            return

        if options["method"] == "mlem":
            self.recs = self._reconstruct_mlem()
            return

        for i in range(self.n_slices):
            self._tmp_sino[:] = self.radios[:, i, :]  # copy into contiguous array
            self.reconstruction.fbp(self._tmp_sino, output=self.recs[i])

    def _reconstruct_cone(self):
        """
        This reconstructs the entire sinograms stack at once
        """

        n_angles, n_z, n_x = self.radios.shape

        # FIXME
        # can't do a discontiguous single copy...
        sinos_contig = self._allocate_array((n_z, n_angles, n_x), np.float32, "sinos_cone")
        for i in range(n_z):
            sinos_contig[i] = self.radios[:, i, :]
        # ---

        # In principle radios are not cropped at this stage,
        # so self.sub_region[2][0] can be used instead of self.get_slice_start_index() instead of self.sub_region[2][0]
        z_min, z_max = self.sub_region_xz[2:]
        n_z_tot = self.process_config.radio_shape(binning=True)[0]

        self.reconstruction.reconstruct(  # pylint: disable=E1101
            sinos_contig,
            output=self.recs,
            relative_z_position=((z_min + z_max) / self.process_config.binning_z / 2) - n_z_tot / 2,
        )

    def _reconstruct_mlem(self):
        """
        This reconstructs the entire sinograms stack at once
        """

        n_angles, n_z, n_x = self.radios.shape

        # FIXME
        # can't do a discontiguous single copy...
        # Initially done for Astra CB recons. But happens that MLEM Corrct also expects
        # data with this order (nb_rows, nb_angles, nb_cols)
        data_vwu = self._allocate_array((n_z, n_angles, n_x), np.float32, "sinos_mlem")
        for i in range(n_z):
            data_vwu[i] = self.radios[:, i, :]
        # ---

        return self.reconstruction.reconstruct(  # pylint: disable=E1101
            data_vwu,
        )

    @pipeline_step("histogram", "Computing histogram")
    def _compute_histogram(self, data=None):
        if data is None:
            data = self.recs
        self.recs_histogram = self.histogram.compute_histogram(data)

    @pipeline_step("writer", "Saving data")
    def _write_data(self, data=None):
        if data is None and self.reconstruction is not None:
            data = self.recs
        if data is None:
            self.logger.info("No data to write")
            return
        self.writer.write_data(data)
        self.logger.info("Wrote %s" % self.writer.fname)
        self._write_histogram()
        self.process_config.single_output_file_initialized = True  # COMPAT.
        self.process_config._writer_initialized = True

    def _write_histogram(self):
        if "histogram" not in self.processing_steps:
            return
        self.logger.info("Saving histogram")
        self.writer.write_histogram(
            hist_as_2Darray(self.recs_histogram),
            processing_index=1,
            config={
                "file": path.basename(self.writer.fname),
                "bins": self.processing_options["histogram"]["histogram_bins"],
            },
        )

    def _process_finalize(self):
        if self.use_margin:
            self.radios = self._orig_radios

    def __repr__(self):
        res = "%s(%s, margin=%s)" % (self.__class__.__name__, str(self.chunk_shape), str(self.margin))
        binning = self.process_config.binning
        subsampling = self.process_config.subsampling_factor

        if binning != (1, 1) or subsampling > 1:
            if binning != (1, 1):
                res += "\nImages binning: %s" % (str(binning))
            if subsampling:
                res += "\nAngles subsampling: %d" % subsampling
            res += "\nRadios chunk: %s ---> %s" % (self.chunk_shape, self.radios_shape)

        if self.use_margin:
            res += "\nMargin: %s" % (str(self.margin))
            res += "\nRadios chunk: %s ---> %s" % (str(self.radios_shape), str(self.radios_cropped_shape))

        res += "\nCurrent subregion: %s" % (str(self.sub_region))
        for step_name in self.processing_steps:
            res += "\n- %s" % (step_name)
        return res

    def _process_chunk(self):
        self._flatfield()
        self._double_flatfield()
        self._ccd_corrections()
        self._rotate_projections()
        self._retrieve_phase()
        self._apply_unsharp()
        self._take_log()
        self._radios_movements()
        self._crop_radios()
        self._normalize_sinos()
        self._destripe_sinos()
        self._dump_sinogram()
        self._reconstruct()
        self._compute_histogram()
        self._write_data()
        self._process_finalize()

    def _reset_reader_subregion(self):
        if self._resume_from_step is not None:
            self.chunk_reader._set_subregion(self.datadump_manager.get_read_dump_subregion())
            self._init_data_dump()
        self._init_reader()

    def _reset_sub_region(self, sub_region):
        self.set_subregion(sub_region)
        self._reset_reader_subregion()
        self._init_flatfield()  # reset flatfield
        self._init_writer()
        self._init_double_flatfield()
        self._init_data_dump()

    def process_chunk(self, sub_region):
        self._reset_sub_region(sub_region)
        self._read_data()
        self._process_chunk()



import numpy
from silx.utils.enum import Enum as _Enum


class SampleSide(_Enum):
    LEFT = "left"
    RIGHT = "right"


class Method(_Enum):
    MEAN = "mean"
    MEDIAN = "median"


def normalize_frame(
    frame: numpy.ndarray, side: SampleSide, method: Method, sample_width: int = 50, margin_before_sample: int = 0
):
    """
    normalize the frame from a sample section picked at the left of the right of the frame

    :param frame: frame to normalize
    :param SampleSide side: side to pick the sample
    :param Method method: normalization method
    :param int sample_width: sample width
    :param int margin: margin before the sampling area
    """
    if not isinstance(frame, numpy.ndarray):
        raise TypeError(f"Frame is expected to be a 2D numpy array.")
    if frame.ndim != 2:
        raise TypeError(f"Frame is expected to be a 2D numpy array. Get {frame.ndim}D")
    side = SampleSide.from_value(side)
    method = Method.from_value(method)

    if frame.shape[1] < sample_width + margin_before_sample:
        raise ValueError(
            f"frame width ({frame.shape[1]}) < sample_width + margin ({sample_width + margin_before_sample})"
        )

    # create sample
    if side is SampleSide.LEFT:
        sample_start = margin_before_sample
        sample_end = margin_before_sample + sample_width
        sample = frame[:, sample_start:sample_end]
    elif side is SampleSide.RIGHT:
        sample_start = frame.shape[1] - (sample_width + margin_before_sample)
        sample_end = frame.shape[1] - margin_before_sample
        sample = frame[:, sample_start:sample_end]
    else:
        raise ValueError(f"side {side.value} not handled")

    # do normalization
    if method is Method.MEAN:
        normalization_array = numpy.mean(sample, axis=1)
    elif method is Method.MEDIAN:
        normalization_array = numpy.median(sample, axis=1)
    else:
        raise ValueError(f"side {side.value} not handled")
    for line in range(normalization_array.shape[0]):
        frame[line, :] -= normalization_array[line]
    return frame



import pytest
from nabu.utils import list_match_queries


def test_list_match_queries():

    # entry0000 .... entry0099
    avail = ["entry%04d" % i for i in range(100)]
    query = "entry0000"
    list_match_queries()



import numpy as np
import pytest
from nabu.reconstruction.rings_cuda import CudaSinoMeanDeringer
from nabu.testutils import compare_arrays, get_data, generate_tests_scenarios, __do_long_tests__
from nabu.reconstruction.rings import MunchDeringer, SinoMeanDeringer, VoDeringer, __has_algotom__
from nabu.thirdparty.pore3d_deringer_munch import munchetal_filter
from nabu.cuda.utils import __has_pycuda__, get_cuda_context

if __has_pycuda__:
    import pycuda.gpuarray as garray
    from nabu.processing.fft_cuda import get_available_fft_implems
    from nabu.thirdparty.tomocupy_remove_stripe import __have_tomocupy_deringer__

    from nabu.reconstruction.rings_cuda import (
        CudaMunchDeringer,
        can_use_cuda_deringer,
        CudaVoDeringer,
    )

    __has_cuda_deringer__ = can_use_cuda_deringer()
else:
    __has_cuda_deringer__ = False
    __have_tomocupy_deringer__ = False

fw_scenarios = generate_tests_scenarios(
    {
        "levels": [4],
        "sigma": [1.0],
        "wname": ["db15"],
        "padding": [(100, 100)],
        "fft_implem": ["vkfft"],
    }
)
if __do_long_tests__:
    fw_scenarios = generate_tests_scenarios(
        {
            "levels": [4, 2],
            "sigma": [1.0, 2.0],
            "wname": ["db15", "haar", "rbio4.4"],
            "padding": [None, (100, 100), (50, 71)],
            "fft_implem": ["skcuda", "vkfft"],
        }
    )


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.sino = get_data("mri_sino500.npz")["data"]
    cls.sino2 = get_data("sino_bamboo_hercules.npz")["data"]
    cls.tol = 5e-3
    cls.rings = {150: 0.5, -150: 0.5}
    if __has_pycuda__:
        cls.ctx = get_cuda_context(cleanup_at_exit=False)
        cls._available_fft_implems = get_available_fft_implems()
    yield
    if __has_pycuda__:
        cls.ctx.pop()


@pytest.mark.usefixtures("bootstrap")
class TestDeringer:
    @staticmethod
    def add_stripes_to_sino(sino, rings_desc):
        """
        Create a new sinogram by adding synthetic stripes to an existing one.

        Parameters
        ----------
        sino: array-like
            Sinogram.
        rings_desc: dict
            Dictionary describing the stripes locations and intensity.
            The location is an integer in [0, N[ where N is the number of columns.
            The intensity is a float: percentage of the current column mean value.
        """
        sino_out = np.copy(sino)
        for loc, intensity in rings_desc.items():
            sino_out[:, loc] += sino[:, loc].mean() * intensity
        return sino_out

    @staticmethod
    def get_fourier_wavelets_reference_result(sino, config):
        # Reference destriping with pore3d "munchetal_filter"
        padding = config.get("padding", None)
        if padding is not None:
            sino = np.pad(sino, ((0, 0), padding), mode="edge")
        ref = munchetal_filter(sino, config["levels"], config["sigma"], wname=config["wname"])
        if config["padding"] is not None:
            ref = ref[:, padding[0] : -padding[1]]
        return ref

    @pytest.mark.skipif(munchetal_filter is None, reason="Need PyWavelets for this test")
    @pytest.mark.parametrize("config", fw_scenarios)
    def test_munch_deringer(self, config):
        deringer = MunchDeringer(
            config["sigma"], self.sino.shape, levels=config["levels"], wname=config["wname"], padding=config["padding"]
        )
        sino = self.add_stripes_to_sino(self.sino, self.rings)
        ref = self.get_fourier_wavelets_reference_result(sino, config)
        # Wrapping with DeRinger
        res = np.zeros((1,) + sino.shape, dtype=np.float32)
        deringer.remove_rings(sino, output=res)

        err_max = np.max(np.abs(res[0] - ref))
        assert err_max < self.tol, "Max error is too high"

    @pytest.mark.skipif(
        not (__has_cuda_deringer__) or munchetal_filter is None,
        reason="Need pycuda, pycudwt and (scikit-cuda or pyvkfft) for this test",
    )
    @pytest.mark.parametrize("config", fw_scenarios)
    def test_cuda_munch_deringer(self, config):
        fft_implem = config["fft_implem"]
        if fft_implem not in self._available_fft_implems:
            pytest.skip("FFT implementation %s is not available" % fft_implem)
        sino = self.add_stripes_to_sino(self.sino, self.rings)
        deringer = CudaMunchDeringer(
            config["sigma"],
            self.sino.shape,
            levels=config["levels"],
            wname=config["wname"],
            padding=config["padding"],
            fft_backend=fft_implem,
            cuda_options={"ctx": self.ctx},
        )
        d_sino = garray.to_gpu(sino)
        deringer.remove_rings(d_sino)
        res = d_sino.get()

        ref = self.get_fourier_wavelets_reference_result(sino, config)

        err_max = np.max(np.abs(res - ref))
        assert err_max < 1e-1, "Max error is too high with configuration %s" % (str(config))

    @pytest.mark.skipif(
        not (__has_algotom__),
        reason="Need algotom for this test",
    )
    def test_vo_deringer(self):
        deringer = VoDeringer(self.sino.shape)
        sino_deringed = deringer.remove_rings_sinogram(self.sino)
        sinos = np.tile(self.sino, (10, 1, 1))
        sinos_deringed = deringer.remove_rings_sinograms(sinos)
        # TODO check result. The generated test sinogram is "too synthetic" for this kind of deringer

    @pytest.mark.skipif(
        not (__have_tomocupy_deringer__),
        reason="Need cupy for this test",
    )
    def test_cuda_vo_deringer(self):
        # Beware, this deringer seems to be buggy for "too-small" sinograms
        # (NaNs on the edges and in some regions). To be investigated

        deringer = CudaVoDeringer(self.sino2.shape)
        d_sino = garray.to_gpu(self.sino2)
        deringer.remove_rings_sinogram(d_sino)
        sino = d_sino.get()

        if __has_algotom__:
            vo_deringer = VoDeringer(self.sino2.shape)
            sino_deringed = vo_deringer.remove_rings_sinogram(self.sino2)

            assert (
                np.max(np.abs(sino - sino_deringed)) < 2e-3
            ), "Cuda implementation of Vo deringer does not yield the same results as base implementation"

    def test_mean_deringer(self):
        deringer_no_filtering = SinoMeanDeringer(self.sino.shape, mode="subtract")

        sino = self.sino.copy()
        deringer_no_filtering.remove_rings_sinogram(sino)

        sino = self.sino.copy()
        deringer_with_filtering = SinoMeanDeringer(self.sino.shape, mode="subtract", filter_cutoff=(0, 30))
        deringer_with_filtering.remove_rings_sinogram(sino)
        # TODO check results

    @pytest.mark.skipif(not (__has_pycuda__), reason="Need pycuda for this test")
    def test_cuda_mean_deringer(self):
        cuda_deringer = CudaSinoMeanDeringer(
            self.sino.shape,
            mode="subtract",
            filter_cutoff=(
                0,
                10,
            ),
            ctx=self.ctx,
        )
        deringer = SinoMeanDeringer(
            self.sino.shape,
            mode="subtract",
            filter_cutoff=(
                0,
                10,
            ),
        )

        d_sino = cuda_deringer.processing.to_device("sino", self.sino)
        cuda_deringer.remove_rings_sinogram(d_sino)

        sino = self.sino.copy()
        sino_d = deringer.remove_rings_sinogram(sino)

        dirac = np.zeros(self.sino.shape[-1], "f")
        dirac[dirac.size // 2] = 1
        deringer_filter_response = deringer._apply_filter(dirac)

        d_dirac = cuda_deringer.processing.to_device("dirac", dirac)
        cuda_deringer_filter_response = cuda_deringer._apply_filter(d_dirac)

        is_close, residual = compare_arrays(
            deringer_filter_response, cuda_deringer_filter_response.get(), 1e-7, return_residual=True
        )
        assert is_close, "Cuda deringer does not have the correct filter response: max_error=%.2e" % residual

        # There is a rather large discrepancy between the vertical_mean kernel and numpy.mean(). Not sure who is right
        is_close, residual = compare_arrays(sino_d, d_sino.get(), 1e-1, return_residual=True)
        assert is_close, (
            "Cuda deringer does not yield the same result as base implementation: max_error=%.2e" % residual
        )



import os
from ..dataset_validator import DatasetValidatorBase


class FullFieldDatasetValidator(DatasetValidatorBase):
    def _validate(self):
        self._check_not_empty()
        self._convert_negative_indices()
        self._get_output_filename()
        self._check_can_do_flatfield()
        self._check_can_do_phase()
        self._check_can_do_reconstruction()
        self._check_slice_indices()
        self._handle_processing_mode()
        self._handle_binning()
        self._check_output_file()

    def _check_can_do_flatfield(self):
        if self.nabu_config["preproc"]["flatfield"]:
            darks = self.dataset_info.darks
            assert len(darks) > 0, "Need at least one dark to perform flat-field correction"
            flats = self.dataset_info.flats
            assert len(flats) > 0, "Need at least one flat to perform flat-field correction"

    def _check_slice_indices(self):
        nx, nz = self.dataset_info.radio_dims
        rec_params = self.rec_params
        if self.is_halftomo:
            ny, nx = self._get_nx_ny()
        what = (("start_x", "end_x", nx), ("start_y", "end_y", nx), ("start_z", "end_z", nz))
        for start_name, end_name, numels in what:
            self._check_start_end_idx(
                rec_params[start_name], rec_params[end_name], numels, start_name=start_name, end_name=end_name
            )

    def _check_can_do_phase(self):
        if self.nabu_config["phase"]["method"] is None:
            return
        self.dataset_info.check_defined_attribute("distance")
        self.dataset_info.check_defined_attribute("pixel_size")

    def _check_can_do_reconstruction(self):
        rec_options = self.nabu_config["reconstruction"]
        if rec_options["method"] is None:
            return
        self.dataset_info.check_defined_attribute("pixel_size")
        if rec_options["method"] == "cone":
            if rec_options["source_sample_dist"] is None:
                err_msg = "In cone-beam reconstruction, you have to provide 'source_sample_dist' in [reconstruction]"
                self.logger.fatal(err_msg)
                raise ValueError(err_msg)
            if rec_options["sample_detector_dist"] is None:
                if self.dataset_info.distance is None:
                    err_msg = "Cone-beam reconstruction: 'sample_detector_dist' was not provided but could not be found in the dataset metadata either. Please provide 'sample_detector_dist'"
                    self.logger.fatal(err_msg)
                    raise ValueError(err_msg)
                self.logger.warning(
                    "Cone-beam reconstruction: 'sample_detector_dist' not provided, will use the one in dataset metadata"
                )
            if self.is_halftomo:
                err_msg = "Cone-beam reconstruction with half-acquisition is not supported yet"
                self.logger.fatal(err_msg)
                raise NotImplementedError(err_msg)



import pytest
import numpy as np
from nabu.utils import median2 as nabu_median_filter
from nabu.testutils import get_data
from nabu.cuda.utils import get_cuda_context, __has_pycuda__
from nabu.preproc.ccd import CCDFilter

if __has_pycuda__:
    import pycuda.gpuarray as garray
    from nabu.preproc.ccd_cuda import CudaCCDFilter


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls

    cls.data = get_data("mri_proj_astra.npz")["data"]
    cls.data /= cls.data.max()
    cls.put_hotspots_in_data()
    if __has_pycuda__:
        cls.ctx = get_cuda_context()


@pytest.mark.usefixtures("bootstrap")
class TestCCDFilter:
    @classmethod
    def put_hotspots_in_data(cls):
        # Put 5 hot spots in the data
        # (row, column, deviation from median)
        cls.hotspots = [(50, 51, 0.04), (151, 150, 0.08), (202, 303, 0.12), (322, 203, 0.14)]
        cls.threshold = 0.1  # parameterize ?
        data_medfilt = nabu_median_filter(cls.data)
        for r, c, deviation_from_median in cls.hotspots:
            cls.data[r, c] = data_medfilt[r, c] + deviation_from_median

    def check_detected_hotspots_locations(self, res):
        diff = self.data - res
        rows, cols = np.where(diff > 0)
        hotspots_arr = np.array(self.hotspots)
        M = hotspots_arr[:, -1] > self.threshold
        hotspots_rows = hotspots_arr[M, 0]
        hotspots_cols = hotspots_arr[M, 1]
        assert np.allclose(hotspots_rows, rows)
        assert np.allclose(hotspots_cols, cols)

    def test_median_clip(self):
        ccd_filter = CCDFilter(self.data.shape, median_clip_thresh=self.threshold)
        res = np.zeros_like(self.data)
        res = ccd_filter.median_clip_correction(self.data, output=res)
        self.check_detected_hotspots_locations(res)

    @pytest.mark.skipif(not (__has_pycuda__), reason="Need cuda/pycuda for this test")
    def test_cuda_median_clip(self):
        d_radios = garray.to_gpu(self.data)
        cuda_ccd_correction = CudaCCDFilter(d_radios.shape, median_clip_thresh=self.threshold)
        d_out = garray.zeros_like(d_radios)
        cuda_ccd_correction.median_clip_correction(d_radios, output=d_out)
        res = d_out.get()
        self.check_detected_hotspots_locations(res)



"""
Base class for CudaKernel and OpenCLKernel
Should not be used directly
"""

from ..utils import updiv


class KernelBase:
    """
    A base class for OpenCL and Cuda kernels.

    Parameters
    -----------
    kernel_name: str
        Name of the CUDA kernel.
    filename: str, optional
        Path to the file name containing kernels definitions
    src: str, optional
        Source code of kernels definitions
    automation_params: dict, optional
        Automation parameters, see below

    Automation parameters
    ----------------------
    automation_params is a dictionary with the following keys and default values.
        guess_block: bool (True)
            If block is not specified during calls, choose a block size based on
            the size/dimensions of the first array.
            Mind that it is unlikely to be the optimal choice.
        guess_grid: bool (True):
            If the grid size is not specified during calls, choose a grid size
            based on the size of the first array.
        follow_device_ptr: bool (True)
            specify gpuarray.gpudata for all cuda GPUArrays (and pyopencl.array.data for pyopencl arrays).
            Otherwise, raise an error.
    """

    _default_automation_params = {
        "guess_block": True,
        "guess_grid": True,
        "follow_device_ptr": True,
    }

    def __init__(
        self,
        kernel_name,
        filename=None,
        src=None,
        automation_params=None,
        silent_compilation_warnings=False,
    ):
        self.check_filename_src(filename, src)
        self.set_automation_params(automation_params)
        self.silent_compilation_warnings = silent_compilation_warnings

    def check_filename_src(self, filename, src):
        err_msg = "Please provide either filename or src"
        if filename is None and src is None:
            raise ValueError(err_msg)
        if filename is not None and src is not None:
            raise ValueError(err_msg)
        if filename is not None:
            with open(filename) as fid:
                src = fid.read()
        self.filename = filename
        self.src = src

    def set_automation_params(self, automation_params):
        self.automation_params = self._default_automation_params.copy()
        self.automation_params.update(automation_params or {})

    @staticmethod
    def guess_grid_size(shape, block_size):
        # python: (z, y, x) -> cuda: (x, y, z)
        res = tuple(map(lambda x: updiv(x[0], x[1]), zip(shape[::-1], block_size)))
        if len(res) == 2:
            res += (1,)
        return res

    @staticmethod
    def guess_block_size(shape):
        """
        Guess a block size based on the shape of an array.
        """
        ndim = len(shape)
        if ndim == 1:
            return (128, 1, 1)
        if ndim == 2:
            return (32, 32, 1)
        else:
            return (16, 8, 8)

    def get_block_grid(self, *args, **kwargs):
        block = None
        grid = None
        if ("block" not in kwargs) or (kwargs["block"] is None):
            if self.automation_params["guess_block"]:
                block = self.guess_block_size(args[0].shape)
            else:
                raise ValueError("Please provide block size")
        else:
            block = kwargs["block"]
        if ("grid" not in kwargs) or (kwargs["grid"] is None):
            if self.automation_params["guess_grid"]:
                grid = self.guess_grid_size(args[0].shape, block)
            else:
                raise ValueError("Please provide block grid")
        else:
            grid = kwargs["grid"]
        self.last_block_size = block
        self.last_grid_size = grid
        return block, grid

    def follow_device_arr(self, args):
        raise ValueError("Base class")

    def _prepare_call(self, *args, **kwargs):
        block, grid = self.get_block_grid(*args, **kwargs)
        # pycuda crashes when any element of block/grid is not a python int (ex. numpy.int64).
        # A weird behavior once observed is "data.shape" returning (np.int64, int, int) (!).
        # Ensure that everything is a python integer.
        grid = tuple(int(x) for x in grid)
        if block is not None:
            block = tuple(int(x) for x in block)
        #
        args = self.follow_device_arr(args)

        return grid, block, args, kwargs



from ..processing.fft_opencl import *
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.opencl.fft has been moved to nabu.processing.fft_opencl", do_print=True, func_name="fft_opencl"
)



import numpy as np
import pytest
from nabu.cuda.utils import get_cuda_context, __has_pycuda__
from nabu.opencl.utils import __has_pyopencl__, get_opencl_context
from nabu.testutils import get_data, generate_tests_scenarios, __do_long_tests__

if __has_pyopencl__:
    from nabu.processing.fftshift import OpenCLFFTshift

configs = {
    "shape": [(300, 451), (300, 300), (255, 300)],
    "axes": [(1,)],
    "dtype_in_out": [(np.float32, np.complex64), (np.complex64, np.float32)],
    "inplace": [True, False],
}

scenarios = generate_tests_scenarios(configs)


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.data = get_data("chelsea.npz")["data"]
    cls.tol = 1e-7
    if __has_pycuda__:
        cls.cu_ctx = get_cuda_context(cleanup_at_exit=False)
    if __has_pyopencl__:
        cls.cl_ctx = get_opencl_context(device_type="all")
    yield
    if __has_pycuda__:
        cls.cu_ctx.pop()


@pytest.mark.skip(reason="OpenCL fftshift is a prototype")
@pytest.mark.usefixtures("bootstrap")
class TestFFTshift:
    def _do_test_fftshift(self, config, fftshift_cls):
        shape = config["shape"]
        dtype = config["dtype_in_out"][0]
        dst_dtype = config["dtype_in_out"][1]
        axes = config["axes"]
        inplace = config["inplace"]
        if inplace and shape[-1] & 1:
            pytest.skip("Not Implemented")
        data = np.ascontiguousarray(self.data[: shape[0], : shape[1]], dtype=dtype)

        backend = fftshift_cls.backend
        ctx = self.cu_ctx if backend == "cuda" else self.cl_ctx
        backend_options = {"ctx": ctx}
        if not (inplace):
            fftshift = fftshift_cls(data.shape, dtype, dst_dtype=dst_dtype, axes=axes, **backend_options)
        else:
            fftshift = fftshift_cls(data.shape, dtype, axes=axes, **backend_options)

        d_data = fftshift.processing.allocate_array("data", shape, dtype)
        d_data.set(data)

        d_res = fftshift.fftshift(d_data)

        assert (
            np.max(np.abs(d_res.get() - np.fft.fftshift(data, axes=axes))) == 0
        ), "something wrong with fftshift_%s(%s)" % (backend, str(config))

    # @pytest.mark.skipif(not (__has_pycuda__), reason="Need pycuda for this test")
    # @pytest.mark.parametrize("config", scenarios)
    # def test_cuda_transpose(self, config):
    # self._do_test_transpose(config, CudaTranspose)

    @pytest.mark.skipif(not (__has_pyopencl__), reason="Need pyopencl for this test")
    @pytest.mark.parametrize("config", scenarios)
    def test_opencl_fftshift(self, config):
        self._do_test_fftshift(config, OpenCLFFTshift)



import numpy as np


try:
    import corrct as cct

    __have_corrct__ = True
except ImportError:
    __have_corrct__ = False


class MLEMReconstructor:
    """
    A reconstructor for MLEM reconstruction using the CorrCT toolbox.
    """

    default_extra_options = {
        "compute_shifts": False,
        "tomo_consistency": False,
        "v_min_for_v_shifts": 0,
        "v_max_for_v_shifts": None,
        "v_min_for_u_shifts": 0,
        "v_max_for_u_shifts": None,
    }

    def __init__(
        self,
        sinos_shape,
        angles_rad,
        shifts_uv=None,
        cor=None,
        n_iterations=50,
        extra_options=None,
    ):
        """ """
        if not (__have_corrct__):
            raise ImportError("Need corrct package")
        self.angles_rad = angles_rad
        self.n_iterations = n_iterations

        self._configure_extra_options(extra_options)
        self._set_sino_shape(sinos_shape)
        self._set_shifts(shifts_uv, cor)

    def _configure_extra_options(self, extra_options):
        self.extra_options = self.default_extra_options.copy()
        self.extra_options.update(extra_options or {})

    def _set_sino_shape(self, sinos_shape):
        if len(sinos_shape) != 3:
            raise ValueError("Expected a 3D shape")
        self.sinos_shape = sinos_shape
        self.n_sinos, self.n_angles, self.prj_width = sinos_shape
        if self.n_angles != len(self.angles_rad):
            raise ValueError(
                f"Number of angles ({len(self.angles_rad)}) does not match size of sinograms ({self.n_angles})."
            )

    def _set_shifts(self, shifts_uv, cor):
        if shifts_uv is None:
            self.shifts_uv = np.zeros([self.n_angles, 2])
        else:
            if shifts_uv.shape[0] != self.n_angles:
                raise ValueError(
                    f"Number of shifts given ({shifts_uv.shape[0]}) does not mathc the number of projections ({self.n_angles})."
                )
            self.shifts_uv = shifts_uv.copy()
        self.cor = cor

    def reconstruct(self, data_vwu):
        """
        data_align_vwu: numpy.ndarray or pycuda.gpuarray
            Raw data, with shape (n_sinograms, n_angles, width)
        output: optional
            Output array. If not provided, a new numpy array is returned
        """
        if not isinstance(data_vwu, np.ndarray):
            data_vwu = data_vwu.get()
        data_vwu /= data_vwu.mean()

        # MLEM recons
        self.vol_geom_align = cct.models.VolumeGeometry.get_default_from_data(data_vwu)
        self.prj_geom_align = cct.models.ProjectionGeometry.get_default_parallel()
        # Vertical shifts were handled in pipeline. Set them to ZERO
        self.shifts_uv[:, 1] = 0.0
        self.prj_geom_align.set_detector_shifts_vu(self.shifts_uv.T[::-1])

        variances_align = cct.processing.compute_variance_poisson(data_vwu)
        self.weights_align = cct.processing.compute_variance_weight(variances_align, normalized=True)  # , use_std=True
        self.data_term_align = cct.data_terms.DataFidelity_wl2(self.weights_align)
        solver = cct.solvers.MLEM(verbose=True, data_term=self.data_term_align)
        self.solver_opts = dict(lower_limit=0)  # , x_mask=cct.processing.circular_mask(vol_geom_align.shape_xyz[:-2])

        with cct.projectors.ProjectorUncorrected(
            self.vol_geom_align, self.angles_rad, rot_axis_shift_pix=self.cor, prj_geom=self.prj_geom_align
        ) as A:
            rec, _ = solver(A, data_vwu, iterations=self.n_iterations, **self.solver_opts)

        return rec



#!/usr/bin/env python
import os
from nabu.pipeline.fullfield.nabu_config import nabu_config


def header(file_):
    content = "# Nabu configuration parameters\nThis file lists all the current configuration parameters available in the [configuration file](nabu_config_file.md)."
    print(content, file=file_)

def generate(file_):
    def write(content):
        print(content, file=file_)
    header(file_)
    for section, values in nabu_config.items():
        if section == "about":
            continue
        write("### %s\n" % section)
        for key, val in values.items():
            if val["type"] == "unsupported":
                continue
            help_content = val["help"]
            if "---" in help_content:
                help_content = help_content.replace("--", "")
            write(help_content + "\n")
            write(
                "```ini\n%s = %s\n```"
                % (key, val["default"])
            )



if __name__ == "__main__":
    fname = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "nabu_config_items.md"
    )
    with open(fname, "w") as f:
        generate(f)



import os
import logging
from typing import Optional, Union
from nabu import version as nabu_version
from nabu.stitching.config import (
    PreProcessedSingleAxisStitchingConfiguration,
    PostProcessedSingleAxisStitchingConfiguration,
    SingleAxisStitchingConfiguration,
)
from nabu.stitching.stitcher.single_axis import PROGRESS_BAR_STITCH_VOL_DESC
from nabu.io.writer import get_datetime
from tomoscan.factory import Factory as TomoscanFactory
from silx.io.dictdump import dicttonx
from nxtomo.application.nxtomo import NXtomo
from tomoscan.utils.volume import concatenate as concatenate_volumes
from tomoscan.esrf.volume import HDF5Volume
from contextlib import AbstractContextManager
from threading import Thread
from time import sleep

_logger = logging.getLogger(__name__)


class StitchingPostProcAggregation:
    """
    for remote stitching each process will stitch a part of the volume or projections.
    Then once all are finished we want to aggregate them all to a final volume or NXtomo.

    This is the goal of this class.
    Please be careful with API. This is already inheriting from a tomwer class

    :param stitching_config: configuration of the stitching configuration
    :param futures: futures that just run
    :param existing_objs: futures that just run
    :param progress_bars: tqdm progress bars for each jobs
    """

    def __init__(
        self,
        stitching_config: SingleAxisStitchingConfiguration,
        futures: Optional[tuple] = None,
        existing_objs_ids: Optional[tuple] = None,
        progress_bars: Optional[dict] = None,
    ) -> None:
        if not isinstance(stitching_config, (SingleAxisStitchingConfiguration)):
            raise TypeError(f"stitching_config should be an instance of {SingleAxisStitchingConfiguration}")
        if not ((existing_objs_ids is None) ^ (futures is None)):
            raise ValueError("Either existing_objs or futures should be provided (can't provide both)")
        if progress_bars is not None and not isinstance(progress_bars, dict):
            raise TypeError(f"'progress_bars' should be None or an instance of a dict. Got {type(progress_bars)}")
        self._futures = futures
        self._stitching_config = stitching_config
        self._existing_objs_ids = existing_objs_ids
        self._progress_bars = progress_bars or {}

    @property
    def futures(self):
        return self._futures

    @property
    def progress_bars(self) -> dict:
        return self._progress_bars

    def retrieve_tomo_objects(self) -> tuple:
        """
        Return tomo objects to be stitched together. Either from future or from existing_objs
        """
        if self._existing_objs_ids is not None:
            scan_ids = self._existing_objs_ids
        else:
            results = {}
            _logger.info(
                f"wait for slurm job to be completed. Advancement will be created once slurm job output file will be available"
            )
            for obj_id, future in self.futures.items():
                results[obj_id] = future.result()

            failed = tuple(
                filter(
                    lambda x: x.exception() is not None,
                    self.futures.values(),
                )
            )
            if len(failed) > 0:
                # if some job failed: useless to do the concatenation
                exceptions = " ; ".join([f"{job} : {job.exception()}" for job in failed])
                raise RuntimeError(f"some job failed. Won't do the concatenation. Exceptiosn are {exceptions}")

            canceled = tuple(
                filter(
                    lambda x: x.cancelled(),
                    self.futures.values(),
                )
            )
            if len(canceled) > 0:
                # if some job canceled: useless to do the concatenation
                raise RuntimeError(f"some job failed. Won't do the concatenation. Jobs are {' ; '.join(canceled)}")
            scan_ids = results.keys()
        return [TomoscanFactory.create_tomo_object_from_identifier(scan_id) for scan_id in scan_ids]

    def dump_stitching_config_as_nx_process(self, file_path: str, data_path: str, overwrite: bool, process_name: str):
        dict_to_dump = {
            process_name: {
                "config": self._stitching_config.to_dict(),
                "program": "nabu-stitching",
                "version": nabu_version,
                "date": get_datetime(),
            },
            f"{process_name}@NX_class": "NXprocess",
        }

        dicttonx(
            dict_to_dump,
            h5file=file_path,
            h5path=data_path,
            update_mode="replace" if overwrite else "add",
            mode="a",
        )

    @property
    def stitching_config(self) -> SingleAxisStitchingConfiguration:
        return self._stitching_config

    def process(self) -> None:
        """
        main function
        """

        # concatenate result
        _logger.info("all job succeeded. Concatenate results")
        if isinstance(self._stitching_config, PreProcessedSingleAxisStitchingConfiguration):
            # 1: case of a pre-processing stitching
            with self.follow_progress():
                scans = self.retrieve_tomo_objects()
            nx_tomos = []
            for scan in scans:
                if not os.path.exists(scan.master_file):
                    raise RuntimeError(
                        f"output file not created ({scan.master_file}). Stitching failed. "
                        "Please check slurm .out files to have more information. Most likely the slurm configuration is invalid. "
                        "(partition name not existing...)"
                    )
                nx_tomos.append(
                    NXtomo().load(
                        file_path=scan.master_file,
                        data_path=scan.entry,
                    )
                )
            final_nx_tomo = NXtomo.concatenate(nx_tomos)
            final_nx_tomo.save(
                file_path=self.stitching_config.output_file_path,
                data_path=self.stitching_config.output_data_path,
                overwrite=self.stitching_config.overwrite_results,
            )

            # dump NXprocess if possible
            parts = self.stitching_config.output_data_path.split("/")
            process_name = parts[-1] + "_stitching"
            if len(parts) < 2:
                data_path = "/"
            else:
                data_path = "/".join(parts[:-1])

            self.dump_stitching_config_as_nx_process(
                file_path=self.stitching_config.output_file_path,
                data_path=data_path,
                process_name=process_name,
                overwrite=self.stitching_config.overwrite_results,
            )

        elif isinstance(self.stitching_config, PostProcessedSingleAxisStitchingConfiguration):
            # 2: case of a post-processing stitching
            with self.follow_progress():
                outputs_sub_volumes = self.retrieve_tomo_objects()
            concatenate_volumes(
                output_volume=self.stitching_config.output_volume,
                volumes=tuple(outputs_sub_volumes),
                axis=1,
            )

            if isinstance(self.stitching_config.output_volume, HDF5Volume):
                parts = self.stitching_config.output_volume.metadata_url.data_path().split("/")
                process_name = parts[-1] + "_stitching"
                if len(parts) < 2:
                    data_path = "/"
                else:
                    data_path = "/".join(parts[:-1])

                self.dump_stitching_config_as_nx_process(
                    file_path=self.stitching_config.output_volume.metadata_url.file_path(),
                    data_path=data_path,
                    process_name=process_name,
                    overwrite=self.stitching_config.overwrite_results,
                )
        else:
            raise TypeError(f"stitching_config type ({type(self.stitching_config)}) not handled")

    def follow_progress(self) -> AbstractContextManager:
        return SlurmStitchingFollowerContext(
            output_files_to_progress_bars={
                job._get_output_file_path(): progress_bar for (job, progress_bar) in self.progress_bars.items()
            }
        )


class SlurmStitchingFollowerContext(AbstractContextManager):
    """Util class to provide user feedback from stitching done on slurm"""

    def __init__(self, output_files_to_progress_bars: dict):
        self._update_thread = SlurmStitchingFollowerThread(file_to_progress_bar=output_files_to_progress_bars)

    def __enter__(self) -> None:
        self._update_thread.start()

    def __exit__(self, *args, **kwargs):
        self._update_thread.join(timeout=1.5)
        for progress_bar in self._update_thread.file_to_progress_bar.values():
            progress_bar.close()  # close to clean display as leave == False


class SlurmStitchingFollowerThread(Thread):
    """
    Thread to check progression of stitching slurm job(s)
    Read slurm jobs .out file each 'delay time' and look for a tqdm line at the end.
    If it exists then deduce progress from it.

    file_to_progress_bar provide for each slurm .out file the progress bar to update
    """

    def __init__(self, file_to_progress_bar: dict, delay_time: float = 0.5) -> None:
        super().__init__()
        self._stop_run = False
        self._wait_time = delay_time
        self._file_to_progress_bar = file_to_progress_bar
        self._first_run = True

    @property
    def file_to_progress_bar(self) -> dict:
        return self._file_to_progress_bar

    def run(self) -> None:
        while not self._stop_run:
            for file_path, progress_bar in self._file_to_progress_bar.items():
                if self._first_run:
                    # make sure each progress bar have been refreshed at least one
                    progress_bar.refresh()

                if not os.path.exists(file_path):
                    continue
                with open(file_path, "r") as f:
                    try:
                        last_line = f.readlines()[-1]
                    except IndexError:
                        continue
                    advancement = self.cast_progress_line_from_log(line=last_line)
                    if advancement is not None:
                        progress_bar.n = advancement
                        progress_bar.refresh()

            self._first_run = False

            sleep(self._wait_time)

    def join(self, timeout: Union[float, None] = None) -> None:
        self._stop_run = True
        return super().join(timeout)

    @staticmethod
    def cast_progress_line_from_log(line: str) -> Optional[float]:
        """Try to retrieve from a line from log the advancement (in percentage)"""
        if PROGRESS_BAR_STITCH_VOL_DESC not in line or "%" not in line:
            return None

        str_before_percentage = line.split("%")[0].split(" ")[-1]
        try:
            advancement = float(str_before_percentage)
        except ValueError:
            _logger.debug(f"Failed to retrieve advancement from log file. Value got is {str_before_percentage}")
            return None
        else:
            return advancement



from ..processing.padding_base import *
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.misc.padding has been moved to nabu.processing.padding_base", do_print=True, func_name="padding_base"
)



import numpy as np
from ..utils import BaseClassError


class _BaseFFT:
    """
    A base class for FFTs.
    """

    implem = "none"
    ProcessingCls = BaseClassError

    def __init__(self, shape, dtype, r2c=True, axes=None, normalize="rescale", **backend_options):
        """
        Base class for Fast Fourier Transform (FFT).

        Parameters
        ----------
        shape: list of int
            Shape of the input data
        dtype: str or numpy.dtype
            Data type of the input data
        r2c: bool, optional
            Whether to use real-to-complex transform for real-valued input. Default is True.
        axes: list of int, optional
            Axes along which FFT is computed.
              * For 2D transform: axes=(1,0)
              * For batched 1D transform of 2D image: axes=(-1,)
        normalize: str, optional
            Whether to normalize FFT and IFFT. Possible values are:
              * "rescale": in this case, Fourier data is divided by "N"
                before IFFT, so that IFFT(FFT(data)) = data.
                This corresponds to numpy norm=None i.e norm="backward".
              * "ortho": in this case, FFT and IFFT are adjoint of eachother,
                the transform is unitary. Both FFT and IFFT are scaled with 1/sqrt(N).
              * "none": no normalizatio is done : IFFT(FFT(data)) = data*N

        Other parameters
        -----------------
        backend_options: dict, optional
            Parameters to pass to CudaProcessing or OpenCLProcessing class.
        """
        self._init_backend(backend_options)
        self._set_dtypes(dtype, r2c)
        self._set_shape_and_axes(shape, axes)
        self._configure_batched_transform()
        self._configure_normalization(normalize)
        self._compute_fft_plans()

    def _init_backend(self, backend_options):
        self.processing = self.ProcessingCls(**backend_options)

    def _set_dtypes(self, dtype, r2c):
        self.dtype = np.dtype(dtype)
        dtypes_mapping = {
            np.dtype("float32"): np.complex64,
            np.dtype("float64"): np.complex128,
            np.dtype("complex64"): np.complex64,
            np.dtype("complex128"): np.complex128,
        }
        if self.dtype not in dtypes_mapping:
            raise ValueError("Invalid input data type: got %s" % self.dtype)
        self.dtype_out = dtypes_mapping[self.dtype]
        self.r2c = r2c

    def _set_shape_and_axes(self, shape, axes):
        # Input shape
        if np.isscalar(shape):
            shape = (shape,)
        self.shape = shape
        # Axes
        default_axes = tuple(range(len(self.shape)))
        if axes is None:
            self.axes = default_axes
        else:
            self.axes = tuple(np.array(default_axes)[np.array(axes)])
        # Output shape
        shape_out = self.shape
        if self.r2c:
            reduced_dim = self.axes[-1] if self.axes is not None else -1
            shape_out = list(shape_out)
            shape_out[reduced_dim] = shape_out[reduced_dim] // 2 + 1
            shape_out = tuple(shape_out)
        self.shape_out = shape_out

    def _configure_batched_transform(self):
        pass

    def _configure_normalization(self, normalize):
        pass

    def _compute_fft_plans(self):
        pass


class _BaseVKFFT(_BaseFFT):
    """
    FFT using VKFFT backend
    """

    implem = "vkfft"
    backend = "none"
    ProcessingCls = BaseClassError
    vkffs_cls = BaseClassError

    def _configure_batched_transform(self):
        if self.axes is not None and len(self.shape) == len(self.axes):
            self.axes = None
            return
        if self.r2c:
            # batched Real-to-complex transforms are supported only along fast axes
            if not (is_fast_axes(len(self.shape), self.axes)):
                raise ValueError("For %dD R2C, only batched transforms along fast axes are allowed" % (len(self.shape)))
            self._vkfft_ndim = len(self.axes)
            self.axes = None  # vkfft still can do a batched transform by providing dim=XX, axes=None

    def _configure_normalization(self, normalize):
        self.normalize = normalize
        self._vkfft_norm = {
            "rescale": 1,
            "backward": 1,
            "ortho": "ortho",
            "none": 0,
        }.get(self.normalize, 1)

    def _set_shape_and_axes(self, shape, axes):
        super()._set_shape_and_axes(shape, axes)
        self._vkfft_ndim = None

    def _compute_fft_plans(self):
        self._vkfft_plan = self.vkffs_cls(
            self.shape,
            self.dtype,
            ndim=self._vkfft_ndim,
            inplace=False,
            norm=self._vkfft_norm,
            r2c=self.r2c,
            dct=False,
            axes=self.axes,
            strides=None,
            **self._vkfft_other_init_kwargs,
        )

    def fft(self, array, output=None):
        if output is None:
            output = self.output_fft = self.processing.allocate_array(
                "output_fft", self.shape_out, dtype=self.dtype_out
            )
        return self._vkfft_plan.fft(array, dest=output)

    def ifft(self, array, output=None):
        if output is None:
            output = self.output_ifft = self.processing.allocate_array("output_ifft", self.shape, dtype=self.dtype)
        return self._vkfft_plan.ifft(array, dest=output)


def is_fast_axes(ndim, axes):
    """
    Return true if "axes" are the fast dimensions
    """
    all_axes = list(range(ndim))
    axes = sorted([ax + ndim if ax < 0 else ax for ax in axes])  # transform "-1" to an actual axis index (1 for 2D)
    return all_axes[-len(axes) :] == axes



import numpy as np
from ..utils import get_opencl_srcfile
from .kernel import OpenCLKernel
from .processing import OpenCLProcessing


class OpenCLMemcpy2D(OpenCLProcessing):
    """
    A class for performing rectangular memory copies between pyopencl arrays.
    It will only work for float32 arrays!
    It was written as pyopencl.enqueue_copy is too cumbersome to use for buffers.
    """

    def __init__(self, ctx=None, device_type="GPU", queue=None, **kwargs):
        super().__init__(ctx=ctx, device_type=device_type, queue=queue, **kwargs)
        self.memcpy2D = OpenCLKernel("cpy2d", self.ctx, filename=get_opencl_srcfile("ElementOp.cl"))

    def __call__(self, dst, src, transfer_shape_xy, dst_offset_xy=None, src_offset_xy=None, wait=True):
        if dst_offset_xy is None:
            dst_offset_xy = (0, 0)
        if src_offset_xy is None:
            src_offset_xy = (0, 0)
        evt = self.memcpy2D(
            self.queue,
            dst,
            src,
            np.int32(dst.shape[-1]),
            np.int32(src.shape[-1]),
            np.int32(dst_offset_xy),
            np.int32(src_offset_xy),
            np.int32(transfer_shape_xy),
            global_size=transfer_shape_xy,
        )
        if wait:
            evt.wait()



from nabu.processing.processing_base import ProcessingBase
from ..utils import deprecated_class

ProcessingBase = deprecated_class("ProcessingBase has been moved to nabu.processing", do_print=True)(ProcessingBase)
